/****************************************************************************
 This is the Automatic Photos module for Roxen WebServer 4.0--
 Copyright (C) 2001--2005 -- M Norrby <norrby+aphoto@lysator.liu.se>

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 To acquire the latest version of aphoto, go to
 <URL:http://www.lysator.liu.se/~norrby/aphoto.html>.

 Please send comments and bug-reports to <norrby+aphoto@lysator.liu.se>.
****************************************************************************/

#include <module.h>
#include <request_trace.h>

inherit "module";

/**
 * Module constants. Read by the server when loading the module.
 */
constant cvs_version = "$Id: aphoto.pike,v 1.184 2005/05/04 16:59:35 norrby Exp $";
int module_type = MODULE_TAG|MODULE_LAST;
constant module_name = "Automatic photos";
constant module_unique = 1;
constant thread_safe = 1;
constant module_doc =
"An image gallery module with automatic thumbnail generation and "
"medium sized images with descriptions for browsing.";


/**
 * Module private constants.
 */
static constant MEDIA_CACHE = "/aphoto_media_cache/";
static constant DIR_STAT_CACHE = "/aphoto_dir_stat_cache/";
static constant DIR_INFO_FILENAME = ".aphoto_dir_info.xml";
static constant DATA_EDITING_PERMISSION = "aphoto editing";
static constant CACHE_TIME = 6912000; //eighty thousand days
static constant EDIT_QUERY = "?edit=true";
static constant MODULE_CREATOR = "norrby+aphoto@lysator.liu.se";
static constant APHOTO = "aphoto";
static constant MODULE_URL = "http://www.lysator.liu.se/~norrby/aphoto.html";
static constant SETTING_BROWSING_IMAGE_MODE = "browsing_image_mode";
static constant SETTING_FOLDER_PAGE = "folder_page";
static constant SETTING_FOLDER_PAGE_URL = "folder_page_url";
static constant SETTING_FOLDER_PAGE_ORIGIN = "folder_page_origin";
static constant SETTING_FOLDER_SORT_ORDER = "folder_sort_order";
static constant SETTING_BROWSING_PAGE = "browsing_page";
static constant SETTING_BROWSING_PAGE_URL = "browsing_page_url";
static constant SETTING_BROWSING_PAGE_ORIGIN = "browsing_page_origin";
static constant SETTING_GALLERY_NAME = "gallery_name";
static constant SETTING_MEDIA_SORT_ORDER = "media_sort_order";
static constant SETTING_THUMBNAIL_MODE = "thumbnail_mode";
static constant SETTING_FONT = "font";
static constant SETTING_FONT_SIZE = "font_size";
static constant SETTING_LETTERS_PER_LINE = "letters_per_line";
static constant SETTING_PATH_SEGMENTS = "path_segments";
static constant SETTING_SUFFIX = "suffix";
static constant SETTING_THUMBNAIL_SIZE = "thumbnail_size";
static constant SETTING_BROWSING_IMAGE_SIZE = "browsing_image_size";
static constant SETTING_USER_AUTH_MODE = "user_auth_mode";
static constant SETTING_UMASK = "umask";
static constant SETTING_SET_UMASK = "set_umask";
static constant SETTING_REDIRECT_SECURE = "redirect-to-secure";
static constant SORT_CRITERIUM_MTIME = "mtime";
static constant SORT_CRITERIUM_NAME = "name";
static constant SORT_CRITERIUM_MTIME_REV = "mtime_reverse";
static constant SORT_CRITERIUM_NAME_REV = "name_reverse";
static constant SSL_CAPABLE_SERVER = "ssl-server";
static constant TAG_APHOTO = "aphoto";
static constant TAG_IMAGE_DESC_SHORT = "image-desc-short";
static constant TAG_IMAGE_DESC_LONG = "image-desc-long";
static constant TAG_IMAGE_POSITION = "image-position";
static constant TAG_IMAGE_DATE = "image-date";
static constant TAG_HIDDEN = "is-hidden";
static constant TAG_DIR_DESC_SHORT = "dir-desc-short";
static constant TAG_DIR_DESC_LONG = "dir-desc-long";
static constant TAG_DIR_DATE = "dir-date";
static constant USER_AUTH_MODE_USERDB = "userdb";
static constant USER_AUTH_MODE_CONFIGIFC = "configifc";
static constant USER_AUTH_MODE_LOCALHOST = "localhost";
static constant USER_AUTH_MODE_FREE = "free";
static constant BROWSING_IMAGE_MODE_SCALED_SHADOWED = "scaled+shadow";
static constant BROWSING_IMAGE_MODE_UNCHANGED = "unchanged";
static constant PAGE_BUILT_IN = "built-in";
static constant PAGE_CONFIG_UI = "config-ui";
static constant PAGE_FROM_URL = "from-url";
static constant THUMBNAIL_MODE_CLASSIC = "classic";
static constant THUMBNAIL_MODE_MEMENTO = "memento";
static constant TYPE_THUMBNAIL = "thumbnail";
static constant TYPE_BROWSING_IMAGE = "browsing_image";
static constant INTERNAL_ARROW_LEFT_NONACTIVE = "arrow-left-nonactive";
static constant INTERNAL_ARROW_LEFT = "arrow-left";
static constant INTERNAL_ARROW_UP_NONACTIVE = "arrow-up-nonactive";
static constant INTERNAL_ARROW_UP = "arrow-up";
static constant INTERNAL_ARROW_RIGHT_NONACTIVE = "arrow-right-nonactive";
static constant INTERNAL_ARROW_RIGHT = "arrow-right";
static constant INTERNAL_FOLDER = "folder";
static constant INTERNAL_CSS = "aphoto.css";
static constant IMAGE_THUMBNAIL_PASSEPARTOUT_SIZE = 4;
static constant IMAGE_BROWSING_IMAGE_PASSEPARTOUT_SIZE = 16;
static constant IMAGE_SHADOW_BORDER_TOPLEFT = 7;
static constant IMAGE_SHADOW_BORDER_BOTTOMRIGHT = 11;
static constant IMAGE_BROWSING_IMAGE_OVERHEAD = IMAGE_SHADOW_BORDER_TOPLEFT
+ 2 * IMAGE_BROWSING_IMAGE_PASSEPARTOUT_SIZE
+ IMAGE_SHADOW_BORDER_BOTTOMRIGHT;
static constant IMAGE_THUMBNAIL_IMAGE_OVERHEAD = IMAGE_SHADOW_BORDER_TOPLEFT
+ 2 * IMAGE_THUMBNAIL_PASSEPARTOUT_SIZE
+ IMAGE_SHADOW_BORDER_BOTTOMRIGHT;

/**
 * Module globals.
 */
roxen.ImageCache image_cache;
Image.Font aphoto_font;
string virtual_server_name;

/**
 * Implemented module API functions in alphabetic order.
 */
void
create(Configuration conf)
{
  set_module_creator(MODULE_CREATOR);
  set_module_url(MODULE_URL);
  roxen.add_permission(DATA_EDITING_PERMISSION, DATA_EDITING_PERMISSION);
  defvar(SETTING_GALLERY_NAME,
	 Variable.String("My gallery",
                         VAR_INITIAL,
                         "Gallery name",
                         "This will be the value returned by &lt;aphoto-gallery"
			 "/&gt;"));
  defvar(SETTING_USER_AUTH_MODE,
	 Variable.StringChoice(initial_user_auth_mode(),
			       ([ USER_AUTH_MODE_USERDB:"System user DB",
                                  USER_AUTH_MODE_CONFIGIFC:"Config ifc user",
				  //USER_AUTH_MODE_LOCALHOST:"localhost",
                                  //USER_AUTH_MODE_FREE:"Free"
			       ]),
			       0,
			       "Auth mode",
                               "Authentication method to user when entering "
                               "edit mode. The <b>System user DB</b> mode "
			       "will only work on the Unix version of Roxen. "
			       "It authenticates the user from the system user "
			       "database. As for the <b>Config ifc user</b> "
			       "mode, a user must be created. Make sure the "
			       "user has the privilege <i>aphoto editing</i>. "
			     /*"The <b>localhost</b> mode will allow everyone "
			       "coming from 127.0.0.1 to edit the album data. "
			       "The <b>Free</b> node will allow everyone to "
			       "edit."*/));
  defvar(SETTING_SET_UMASK,
         Variable.Flag(0,
                       0,
                       "Change umask",
                       "If the server is to set the umask before creating new "
                       "files such as temporary files and directory info. This "
                       "setting is useful if more than one person is editing "
                       "the files, and a common group is used for "
                       "write permissions."));
  defvar(SETTING_UMASK,
         Variable.String("002",
                         0,
                         "Umask",
                         "Value of the umask. The value 002 "
                         "means full permissions for both user and group."));

  getvar(SETTING_UMASK)->
    set_invisibility_check_callback(lambda(RequestID, Variable.Variable)
                                    {
                                      return !query(SETTING_SET_UMASK);
                                    });
  getvar(SETTING_UMASK)->
    set_changed_callback(lambda(Variable.Variable v)
                         {
                           int octvalue;
                           string empty;
                           int matches = sscanf((string)v->query(),
                                                "%o%s",
                                                octvalue,
                                                empty);
                           if (matches != 2 || empty != "")
                             v->low_set(v->default_value());
                         });
  defvar(SETTING_REDIRECT_SECURE,
         Variable.Flag(0,
                       0,
                       "Redirect editing",
                       "If set, editing requests will be redirected to another "
		       "server if the current server is not SSL capable. "
		       "If you are not careful, this can result in "
		       "an eternal redirection loop if target server is not "
		       "SSL capable either. After saving this variable, "
		       "another setting will appear, asking you to supply "
		       "the secure server."));
  defvar(SSL_CAPABLE_SERVER,
	 Variable.String("https://example.com",
                         0,
                         "Secure server",
                         "The content of this parameter will be used for "
			 "secure server redirection."));
  getvar(SSL_CAPABLE_SERVER)->
    set_changed_callback(lambda(Variable.Variable v)
                         {
                           string server = (string) v->query();
			   if (!!server && server != "" && server[-1] == '/')
			     v->low_set(server[..sizeof(server) - 2]);
                         });
  getvar(SSL_CAPABLE_SERVER)->
    set_invisibility_check_callback(lambda(RequestID, Variable.Variable)
                                    {
                                      return !query(SETTING_REDIRECT_SECURE);
                                    });
  defvar(SETTING_FOLDER_PAGE_ORIGIN,
	 Variable.StringChoice(PAGE_BUILT_IN,
			       ([ PAGE_BUILT_IN:"Built in",
                                  PAGE_CONFIG_UI:"Define as variable in this "
				  "UI",
                                  PAGE_FROM_URL:"Fetch from file under "
                                  "this virtual server",
			       ]),
			       0,
			       "Folder page",
                               "Where to find the folder page template "
                               "containing RXML code for thumbnails and "
                               "sub folders. If you chose to define the "
			       "page in this UI, you have to press save "
			       "for an edit box to appear!"));
  defvar(SETTING_FOLDER_PAGE,
	 Variable.Text(DEFAULT_FOLDER_PAGE,
		       0,
                       "Edit folder page"
                       ""));
  getvar(SETTING_FOLDER_PAGE)->
    set_invisibility_check_callback(lambda(RequestID, Variable.Variable)
                                    {
                                      return query(SETTING_FOLDER_PAGE_ORIGIN)
                                        != PAGE_CONFIG_UI;
                                    });
  defvar(SETTING_FOLDER_PAGE_URL,
	 Variable.String("/complete/virtual/path/to/template.rxml",
                         0,
                         "Path to host relative folder page RXML code.",
                         "Path under the current host where a folder page "
                         "template file can be found. If the file resides "
                         "at http://your.host.example.com/foo/page.rxml, "
                         "the \"http://your.host.example.com\" part should "
                         "be left out."));
  getvar(SETTING_FOLDER_PAGE_URL)->
    set_invisibility_check_callback(lambda(RequestID, Variable.Variable)
                                    {
                                      return query(SETTING_FOLDER_PAGE_ORIGIN)
                                        != PAGE_FROM_URL;
                                    });
  defvar(SETTING_THUMBNAIL_SIZE,
	 Variable.Int(128,
		      0,
		      "Thumbnail size",
		      "Length in pixels of generated thumbnails."));

  defvar(SETTING_THUMBNAIL_MODE,
	 Variable.StringChoice(THUMBNAIL_MODE_MEMENTO,
			       ([ THUMBNAIL_MODE_MEMENTO: "Memento style "
                                  "(with descriptions on thumbnail images)",
                                  THUMBNAIL_MODE_CLASSIC: "Classic style "
                                  "(without generated text on thumbnail images)"
			       ]),
			       0,
			       "Thumbnail viewing mode",
                               ""));
  defvar(SETTING_BROWSING_IMAGE_MODE,
	 Variable.StringChoice(BROWSING_IMAGE_MODE_SCALED_SHADOWED,
			       ([ BROWSING_IMAGE_MODE_SCALED_SHADOWED: "Scaled"
                                  " (with backdrop shadow)",
                                  BROWSING_IMAGE_MODE_UNCHANGED: "Unchanged"
                                  " (without resizing or dropshadows))"
			       ]),
			       0,
			       "Browsing image mode",
                               "Scaled, medium sized images, with backdrop "
                               "shadows can be generated for nicer and more "
                               "convenient browsing. This is the default."));
  defvar(SETTING_BROWSING_PAGE_ORIGIN,
	 Variable.StringChoice(PAGE_BUILT_IN,
			       ([ PAGE_BUILT_IN:"Built in",
                                  PAGE_CONFIG_UI:"Define as variable in this "
				  "UI",
                                  PAGE_FROM_URL:"Fetch from file under "
                                  "this virtual server",
			       ]),
			       0,
			       "Browsing page",
                               "Where to find the browsing page template "
                               "containing RXML code. If you chose to define "
			       "the page in this UI, you have to press save "
			       "for an edit box to appear!"));
  defvar(SETTING_BROWSING_PAGE,
	 Variable.Text(DEFAULT_BROWSING_PAGE,
		       0,
                       "Edit browsing page"
                       ""));

  getvar(SETTING_BROWSING_PAGE)->
    set_invisibility_check_callback(lambda(RequestID, Variable.Variable)
                                    {
                                      return query(SETTING_BROWSING_PAGE_ORIGIN)
                                        != PAGE_CONFIG_UI;
                                    });
  defvar(SETTING_BROWSING_PAGE_URL,
	 Variable.String("/complete/virtual/path/to/template.rxml",
                         0,
                         "Path to host relative browsing page RXML code.",
                         "Path under the current host where a browsing page "
                         "template file can be found. If the file resides "
                         "at http://your.host.example.com/foo/page.rxml, "
                         "the \"http://your.host.example.com\" part should "
                         "be left out."));
  getvar(SETTING_BROWSING_PAGE_URL)->
    set_invisibility_check_callback(lambda(RequestID, Variable.Variable)
                                    {
                                      return query(SETTING_BROWSING_PAGE_ORIGIN)
                                        != PAGE_FROM_URL;
                                    });
  defvar(SETTING_BROWSING_IMAGE_SIZE,
	 Variable.Int(640,
		      0,
		      "Browsing image size",
		      "Size in pixels of generated browsing images."));

  getvar(SETTING_BROWSING_IMAGE_SIZE)->
    set_invisibility_check_callback(lambda(RequestID, Variable.Variable)
                                    {
                                      return query(SETTING_BROWSING_IMAGE_MODE)
                                        != BROWSING_IMAGE_MODE_SCALED_SHADOWED;
                                    });

  defvar(SETTING_PATH_SEGMENTS,
	 Variable.Int(3,
		      0,
		      "Path segments",
		      "When displaying paths as titles on folder pages, "
                      "the path can be very long, causing the nonbreakable "
                      "H1 element to widen the page very much. Using this "
                      "setting will make /long/path/foo/bar become "
                      ".../foo/bar."));
  defvar(SETTING_SUFFIX,
	 Variable.String(".aphoto",
                         0,
                         "Suffix",
                         "Appending this suffix to a path will cause "
                         "aphoto to handle the request."));
  defvar(SETTING_FONT,
	 Variable.FontChoice("julius thyssen",
                             0,
                             "Font",
                             "When in memento mode, thumbnails have a small "
                             "test on the side or in the bottom. This "
                             "settings controls which font to use."));
  getvar(SETTING_FONT)->
    set_invisibility_check_callback(memento_mode_invisibility_callback);

  defvar(SETTING_FONT_SIZE,
	 Variable.Int(17,
		      0,
		      "Font size",
		      "Size of font used when generating "
                      "short image descriptions when in memento mode."));
  getvar(SETTING_FONT_SIZE)->
    set_invisibility_check_callback(memento_mode_invisibility_callback);

  defvar(SETTING_LETTERS_PER_LINE,
	 Variable.Int(18,
		      0,
		      "Letters per line",
		      "Maximum number of letters per line "
                      "in a generated short description. "
                      "Used only in memento mode."));
  getvar(SETTING_LETTERS_PER_LINE)->
    set_invisibility_check_callback(memento_mode_invisibility_callback);

  defvar(SETTING_MEDIA_SORT_ORDER,
	 Variable.StringChoice(SORT_CRITERIUM_MTIME,
			       ([ SORT_CRITERIUM_MTIME:"Time",
                                  SORT_CRITERIUM_MTIME_REV:"Time, reverse",
				  SORT_CRITERIUM_NAME:"Filename",
				  SORT_CRITERIUM_NAME_REV:"Filename, reverse"
			       ]),
			       0,
			       "Image sort order",
                               "Thumbnails are sorted like this, and "
                               "browsing arrows for images are also "
                               "affected."));

  defvar(SETTING_FOLDER_SORT_ORDER,
	 Variable.StringChoice(SORT_CRITERIUM_MTIME,
			       ([ SORT_CRITERIUM_MTIME:"Time",
                                  SORT_CRITERIUM_MTIME_REV:"Time, reverse",
				  SORT_CRITERIUM_NAME:"Foldername",
				  SORT_CRITERIUM_NAME_REV:"Foldername, reverse"
			       ]),
			       0,
			       "Folder sort order",
                               ""));
}

mapping|int(-1..0)|Stdio.File
find_internal(string file, RequestID id)
{
  switch (basename(file)) {
  case INTERNAL_CSS:
    return Roxen.http_string_answer(APHOTO_CSS, "text/css");
    break;
  case INTERNAL_FOLDER:
    return Roxen.http_string_answer(APHOTO_FOLDER, "image/gif");
    break;
  case INTERNAL_ARROW_RIGHT:
    return Roxen.http_string_answer(ARROW_RIGHT, "image/gif");
    break;
  case INTERNAL_ARROW_UP:
    return Roxen.http_string_answer(ARROW_UP, "image/gif");
    break;
  case INTERNAL_ARROW_LEFT:
    return Roxen.http_string_answer(ARROW_LEFT, "image/gif");
    break;
  case INTERNAL_ARROW_RIGHT_NONACTIVE:
    return Roxen.http_string_answer(ARROW_RIGHT_NONACTIVE, "image/gif");
    break;
  case INTERNAL_ARROW_UP_NONACTIVE:
    return Roxen.http_string_answer(ARROW_UP_NONACTIVE, "image/gif");
    break;
  case INTERNAL_ARROW_LEFT_NONACTIVE:
    return Roxen.http_string_answer(ARROW_LEFT_NONACTIVE, "image/gif");
    break;
  }
  return image_cache->http_file_answer((file/".")[0], id);
}

mapping|int
last_resort(RequestID id)
{
  if (is_my_request(id))
    return handle_request(id);
  else
    return 0;
}

mapping
parse_directory(RequestID id) {
  return generate_folder_page(id);
}

mapping(LocaleString:function(RequestID:void))
query_action_buttons(RequestID id)
{
  return (["Clear image cache":flush_cache]);
}

void
start(int num, Configuration config)
{
  image_cache = roxen.ImageCache("aphoto", generate_image);

  array dependencies = ({"contenttypes",
			 "rxmlparse",
			 "rxmltags",
                         "auth_httpbasic",
                         "userdb_system"});

  module_dependencies(config, dependencies);
  aphoto_font = get_font(query(SETTING_FONT),
                         query(SETTING_FONT_SIZE),
                         0,
                         0,
                         "left",
                         1.0,
                         1.0);
  virtual_server_name = config->query_name();
}

string
status()
{
  array image_cache_status = image_cache->status();
  return
    "<b>Current image cache size:</b> "
    + Roxen.sizetostring(image_cache_status[1])
    + "</p><p>"
    + (supports_jpeg_transforms() ? "" : "<b>NOTE!</b> Sorry, <b>no</b>")
    + " JPEG image"
    + " transforms supported by current Pike version."
    + (supports_jpeg_transforms() ? "" : " The aphoto module"
       " will thus not let you edit image texts or rotate images.")
    + "</p>";
}

/**
 * Aphoto: internal functions in alphabetic order.
 */

static Image.image
add_dropshadow(Image.Image in,
               int border1,
               int border2)
{
  Image.image bg = Image.image(in->xsize() + border1 + border2,
			       in->ysize() + border1 + border2,
			       0x60,
                               0x60,
                               0x60);
  int cs = corner1->xsize();
  int cw = bg->xsize() - 2 * cs;
  int ch = bg->ysize() - 2 * cs;
  bg->paste(corner1, 0, 0);
  bg->paste(corner2, bg->xsize() - cs, 0);
  bg->paste(corner3, bg->xsize() - cs, bg->ysize() - cs);
  bg->paste(corner4, 0, bg->ysize() - cs);
  bg->paste(bg->copy(cs - 1, 0, cs - 1, cs - 1)->scale(cw, cs), cs, 0);
  bg->paste(bg->copy(cs - 1,
                     ch + cs,
                     cs - 1,
                     ch + cs + cs - 1)->scale(cw, cs), cs, ch + cs);
  bg->paste(bg->copy(0, cs - 1, cs - 1, cs - 1)->scale(cs, ch), 0, cs);
  bg->paste(bg->copy(cw + cs - 1, cs - 1, cw + cs + cs - 1, cs - 1)->scale(cs,
                                                                           ch),
	    cw+cs, cs);
  bg->paste(in, border1, border1);
  return bg;
}

static mapping
add_slash_relocation(RequestID id)
{
  string relocation = get_aphoto_file(id) + "/"  + get_aphoto_folder_suffix(id);
  if (get_aphoto_query(id)) relocation += get_aphoto_query(id);
  return Roxen.http_redirect(relocation, id);
}

static int(0..1)
aphoto_has_folder_suffices(RequestID id)
{
  if (!id->misc->aphoto_has_folder_suffices)
    id->misc->aphoto_has_folder_suffices =
      !id->conf->find_module("aphoto_directories");
  return id->misc->aphoto_has_folder_suffices;
}

static Stdio.Stat
aphoto_stat_file(RequestID id)
{
  if (!id->misc->aphoto_stat_file) {
    id->misc->aphoto_stat_file = file_stat(get_aphoto_real_file(id));
  }
  return id->misc->aphoto_stat_file;
}

static mapping
browsing_page_posted_redirect(RequestID id)
{
  if (!!id->variables->submit_and_previous
      && get_aphoto_previous_browsing_page(id)) {
    return Roxen.http_redirect(get_aphoto_previous_browsing_page(id)
                               + get_aphoto_query(id), id);
  }
  if (!!id->variables->submit_and_next
      || !!id->variables->hide_and_next
      || !!id->variables->unhide) {
    if (get_aphoto_next_browsing_page(id))
      return Roxen.http_redirect(get_aphoto_next_browsing_page(id)
                                 + get_aphoto_query(id), id);
    else
      return Roxen.http_redirect(get_aphoto_this_browsing_page(id), id);
  }
  return Roxen.http_redirect(get_aphoto_this_browsing_page(id), id);
}

static int(0..)
caption_pos(int photo_width, int caption_width)
{
  return max(0, (photo_width - caption_width)/2);
}

static array(int(0..))
complete_thumbnail_size(int(0..) width, int(0..) height)
{
  return ({width + IMAGE_THUMBNAIL_IMAGE_OVERHEAD,
           height + IMAGE_THUMBNAIL_IMAGE_OVERHEAD});
}

static mapping(string:Stdio.Stat)
array_to_stats(mapping(string:array(int)) stats)
{
  if (!!stats)
    return mkmapping(indices(stats), map(values(stats), Stdio.Stat));
  return 0;
}

static Medium
create_medium(string filename, Stdio.Stat stat, RequestID id)
{
  if (is_jpeg_name(filename, id)) {
    return JpegImageMedium(filename, stat, id);
  }
  if (is_image_name(filename, id)) {
    return ImageMedium(filename, stat, id);
  }
  if (is_movie_name(filename, id)) {
    return MovieMedium(filename, stat, id);
  }
  if (is_dir(stat)) {
    return AphotoDir(filename, stat, id);
  }
}

static int(0..1)
dimensions_equal_or_rotated(string old_image, string new_image)
{
  mapping old_header = Image.JPEG.decode_header(old_image);
  mapping new_header = Image.JPEG.decode_header(new_image);
  return (equal(sort(({old_header->xsize, old_header->ysize})),
                sort(({old_header->xsize, old_header->ysize}))));
}

static int(0..1)
dir_changed_since_last_time(mapping(string:Stdio.Stat) stats, RequestID id)
{
  if (!stats) return 1;
  mapping(string:Stdio.Stat) old_stats = get_aphoto_dir_stat_cache(id);
  if (!old_stats || sizeof(stats) != sizeof(old_stats)) {
    set_aphoto_dir_stat_cache(stats, id);
    return 1;
  }
  foreach (indices(old_stats), string stat)
    if (!stats[stat]
        || !old_stats[stat]
        || stats[stat]->mtime != old_stats[stat]->mtime) {
      set_aphoto_dir_stat_cache(stats, id);
      return 1;
    }
  return 0;
}

static int(0..1)
editing_data_posted(RequestID id)
{
  return !!id->variables->aphoto_editing_data &&
    id->variables->aphoto_editing_data == "posted";
}

static mapping
error_answer(string message, RequestID id) {
  return Roxen.http_string_answer(message, "text/plain");
}

static mapping
extract_jpeg_header(string raw_jpeg_file)
{
  return Image.JPEG.decode_header(raw_jpeg_file);
}

static array(mixed)
extract_aphoto_image_xml(string aphoto_data)
{
  string top_tag_data = extract_xml(TAG_APHOTO, aphoto_data);
  string pos_xml = extract_xml(TAG_IMAGE_POSITION, top_tag_data);
  array(string) positions;
  if (pos_xml) {
    positions = (pos_xml/"-");
  }
  return
    ({extract_xml(TAG_IMAGE_DESC_SHORT, top_tag_data),
      extract_xml(TAG_IMAGE_DESC_LONG, top_tag_data),
      ((int) extract_xml(TAG_IMAGE_DATE, top_tag_data)),
      !!extract_xml(TAG_HIDDEN, top_tag_data),
      positions && sizeof(positions) > 0 ? (int) positions[0] : 0,
      positions && sizeof(positions) > 1 ? (int) positions[1] : 0,
      positions && sizeof(positions) > 2 ? (int) positions[2] : 0
    });
}

static array(mixed)
extract_aphoto_folder_xml(string aphoto_data)
{
  string top_tag_data = extract_xml(TAG_APHOTO, aphoto_data);
  return
    ({extract_xml(TAG_DIR_DESC_SHORT, top_tag_data),
      extract_xml(TAG_DIR_DESC_LONG, top_tag_data),
      ((int) extract_xml(TAG_DIR_DATE, top_tag_data)),
      !!extract_xml(TAG_HIDDEN, top_tag_data)
    });
}

static string|int(0..1)
extract_xml(string tag, string text_block)
{
  if (!text_block) return 0;
  if (sscanf(text_block, "%*s<"+tag+"%*s>%s</"+tag+">",
	     text_block) != 3)
    return sscanf(text_block, "%*s<"+tag+"%*s/>") >= 1;
  return text_block;
}

static int
exif_time(string real_path)
{
  mapping exif;
  catch {
    //werror("parsing EXIF data on %s.\n", real_path);
    exif = Standards.EXIF.get_properties(Stdio.FILE(real_path));
    //werror("parsing EXIF data OK.\n");
  };
  //werror("After EXIF data.\n");
  if (!!exif && sizeof(exif)) {
    if (exif->DateTime)
      return unixtime(exif->DateTime);
  }
  else {
    //werror("No good EXIF data found!\n");
  }
  return 0;
}

static void
flush_cache()
{
  image_cache->flush();
}

static Image.Image
generate_browsing_image(string raw_image, string filename, RequestID id)
{
  int unused;
  [string title, string long_desc, int date, int hidden, unused, unused, unused]
    = get_image_data(raw_image);
  Image.Image image = make_resized_image(raw_image,
                                         query(SETTING_BROWSING_IMAGE_SIZE));
  image = make_border(image, IMAGE_BROWSING_IMAGE_PASSEPARTOUT_SIZE);
  return add_dropshadow(image,
                        IMAGE_SHADOW_BORDER_TOPLEFT,
                        IMAGE_SHADOW_BORDER_BOTTOMRIGHT);
}

static mapping
generate_browsing_page(RequestID id)
{
  if (in_edit_mode(id) && insecure(id) && should_redirect_to_ssl(id))
    return redirect_to_secure_server(id);
  if (!get_aphoto_editor(id)->authenticated())
    return id->conf->authenticate_throw(id, DATA_EDITING_PERMISSION);
  if (editing_data_posted(id)) {
    return handle_browsing_page_editing_post(id);
  }
  Medium medium = get_aphoto_medium(id);
  if (in_edit_mode(id) && is_editable_data(id))
    return Roxen.http_rxml_answer(DEFAULT_BROWSING_EDIT_PAGE, id);
  return Roxen.http_rxml_answer(get_browsing_page(id), id);
}

static mapping
generate_folder_page(RequestID id)
{
  if (in_edit_mode(id) && insecure(id) && should_redirect_to_ssl(id))
    return redirect_to_secure_server(id);
  if (!get_aphoto_editor(id)->authenticated()) {
    return id->conf->authenticate_throw(id, DATA_EDITING_PERMISSION);
  }
  if (editing_data_posted(id)) {
    return handle_folder_editing_post(id);
  }
  Medium dir = get_aphoto_medium(id);
  if (in_edit_mode(id))
    return Roxen.http_rxml_answer(DEFAULT_FOLDER_EDIT_PAGE, id);
  return Roxen.http_rxml_answer(get_folder_page(id), id);
}

static array(Image.Layer)
generate_image(mapping arg, RequestID id)
{
  array layers = ({});
  mapping opts = ([]);
  string raw_image = id->conf->try_get_file(arg->file, id);
  Image.Image ret;
  if (arg->type == TYPE_THUMBNAIL) {
    ret = generate_thumbnail(raw_image, basename(arg->file), id);
  }
  else if (arg->type == TYPE_BROWSING_IMAGE) {
    ret = generate_browsing_image(raw_image, basename(arg->file), id);
  }
  if (!!arg->hidden)
    ret = ret->grey(0,255, 0);
  return ({Image.Layer(ret)});
}

Image.Image
generate_thumbnail(string raw_image, string filename, RequestID id)
{
  int unused;
  int(0..1) hidden;
  [string title, string not_used_string, unused, hidden, unused, unused, unused]
    = get_image_data(raw_image);

  Image.Image img = make_resized_image(raw_image,
                                       get_aphoto_thumbnail_size(id));

  [int xsz, int ysz] = get_thumbnail_size_from_image(title, filename, img, id);
  Image.Image photo = Image.Image(xsz + 2 * IMAGE_THUMBNAIL_PASSEPARTOUT_SIZE,
                                  ysz + 2 * IMAGE_THUMBNAIL_PASSEPARTOUT_SIZE,
                                  Image.Color.white);
  photo->paste(img,
               IMAGE_THUMBNAIL_PASSEPARTOUT_SIZE,
               IMAGE_THUMBNAIL_PASSEPARTOUT_SIZE);

  if (hidden || should_make_title_field_on_thumbnail(title, filename, id)) {
    Image.Image caption
      = aphoto_font->write(@split_title_in_lines(title, hidden));
    Image.Image background = Image.Image(xsz,
                                         caption->ysize() + 6,
                                         Image.Color.white);
    caption = background->paste_alpha_color(caption,
                                            Image.Color.blue,
                                            caption_pos(background->xsize(),
                                                        caption->xsize()),
                                            2);
    if (is_horizontal(img))
      photo->paste(caption,
                   IMAGE_THUMBNAIL_PASSEPARTOUT_SIZE,
                   min(img->ysize(), photo->ysize() - caption->ysize()));
    else
      photo->paste(caption->rotate_ccw(),
                   min(img->xsize(), photo->xsize() - caption->ysize()),
                   IMAGE_THUMBNAIL_PASSEPARTOUT_SIZE);
  }
  return add_dropshadow(photo,
                        IMAGE_SHADOW_BORDER_TOPLEFT,
                        IMAGE_SHADOW_BORDER_BOTTOMRIGHT);
}

static string
get_aphoto_basename(RequestID id)
{
  if (!id->misc->aphoto_basename)
    id->misc->aphoto_basename = basename(get_aphoto_file(id));
  return id->misc->aphoto_basename;
}

static array(string)
get_aphoto_browsing_image_size(RequestID id)
{
  if (!id->misc->aphoto_browsing_image_size) {
    [int X, int Y] = get_aphoto_raw_image_size(id);
    float scale = min(((float) query(SETTING_BROWSING_IMAGE_SIZE) / max(X, Y)),
                      1.0);
    int x = IMAGE_BROWSING_IMAGE_OVERHEAD + (int) (X * scale);
    int y = IMAGE_BROWSING_IMAGE_OVERHEAD + (int) (Y * scale);
    id->misc->aphoto_browsing_image_size = ({(string) x, (string) y});
  }
  return id->misc->aphoto_browsing_image_size;
}

static string
get_aphoto_cache_key(string which_cache, string set, string set2, RequestID id)
{
  string long_key = virtual_server_name + which_cache + query(set)
    + ":" + query(set2) + get_aphoto_virtdir(id);
  return Crypto.md5()->update(long_key)->digest();
}

static mapping(string:Stdio.Stat)
get_aphoto_dir_stat(RequestID id)
{
  if (!id->misc->aphoto_dir_stat)
    id->misc->aphoto_dir_stat = id->conf->find_dir_stat(get_aphoto_virtdir(id),
                                                        id) || ([]);
  return id->misc->aphoto_dir_stat;
}

static mapping(string:Stdio.Stat)
get_aphoto_dir_stat_cache(RequestID id)
{
  string key = get_aphoto_cache_key(DIR_STAT_CACHE,
				    SETTING_MEDIA_SORT_ORDER,
				    SETTING_THUMBNAIL_MODE,
				    id);
  string cached_data = cache.get_session_data(key);
  if (!!cached_data)
    return array_to_stats(decode_value(cached_data));
  return ([]);
}

static AlbumEditor
get_aphoto_editor(RequestID id) {
  if (!id->misc->aphoto_editor) {
    AlbumEditor editor;
    switch (query(SETTING_USER_AUTH_MODE)) {
    case USER_AUTH_MODE_USERDB:
      editor = UserDBEditor(id);
      break;
    case USER_AUTH_MODE_CONFIGIFC:
      editor = ConfigUIEditor(id);
      break;
    default:
      werror("Unknown user auth mode \"%s\"", query(SETTING_USER_AUTH_MODE));
    }
    id->misc->aphoto_editor = editor;
  }
  return id->misc->aphoto_editor;
}

array(AphotoDir)
get_aphoto_folders(RequestID id)
{
  if (!id->misc->aphoto_folders) {
    mapping(string:Stdio.Stat) stats = get_aphoto_dir_stat(id);
    array(AphotoDir) folders = ({});
    foreach (indices(stats), string filename) {
      if (stats[filename]->isdir) {
        folders += ({AphotoDir(filename, stats[filename], id)});
      }
    }
    id->misc->aphoto_folders
      = Array.sort_array(folders, get_sort_fun(SETTING_FOLDER_SORT_ORDER));
  }
  return id->misc->aphoto_folders;
}

static string
get_aphoto_file(RequestID id)
{
  if (!id->misc->aphoto_file) {
    if (has_suffix(id->not_query, get_aphoto_suffix(id)))
      id->misc->aphoto_file = id->not_query[0..sizeof(id->not_query)
                                            - sizeof(get_aphoto_suffix(id))
                                            - 1];
    else
      id->misc->aphoto_file = id->not_query;
  }
  return id->misc->aphoto_file;
}

static string
get_aphoto_folder_suffix(RequestID id)
{
  if (!id->misc->aphoto_folder_suffix)
    if (aphoto_has_folder_suffices(id))
      id->misc->aphoto_folder_suffix = get_aphoto_suffix(id);
    else
      id->misc->aphoto_folder_suffix = "";
  return id->misc->aphoto_folder_suffix;
}

static string
get_aphoto_gallery_name(RequestID id)
{
  if (!id->misc->aphoto_gallery_name) {
    id->misc->aphoto_gallery_name = query(SETTING_GALLERY_NAME);
  }
  return id->misc->aphoto_gallery_name;
}

static int
get_aphoto_image_transform(RequestID id)
{
  int transform;
  if (supports_jpeg_transforms()) {
    switch (id->variables->trans) {
#if constant(Image.JPEG.ROT_90)
    case "rot90":
      transform = Image.JPEG.ROT_90;
      break;
    case "rot180":
      transform = Image.JPEG.ROT_180;
      break;
    case "rot270":
      transform = Image.JPEG.ROT_270;
      break;
    case "fliplr":
      transform = Image.JPEG.FLIP_H;
      break;
    case "flipud":
      transform = Image.JPEG.FLIP_V;
      break;
    case "nop":
      int transform = Image.JPEG.NONE;
      break;
#endif
    default:
      werror("Transform argument missing!");
#if constant(Image.JPEG.ROT_90)
      return Image.JPEG.NONE;
#endif
    }
  }
  return transform;
}

static string
get_aphoto_internal_location(RequestID id)
{
  if (!id->misc->aphoto_internal_location)
    id->misc->aphoto_internal_location = query_absolute_internal_location(id);
  return id->misc->aphoto_internal_location;
}

static int(0..)
get_aphoto_maximum_path_segments(RequestID id)
{
  if (!id->misc->aphoto_maximum_path_segments)
    id->misc->aphoto_maximum_path_segments = query(SETTING_PATH_SEGMENTS);
  return id->misc->aphoto_maximum_path_segments;
}

static array(Medium)
get_aphoto_media_cache(RequestID id)
{
  string cached_data =
    cache.get_session_data(get_aphoto_cache_key(MEDIA_CACHE,
                                                SETTING_MEDIA_SORT_ORDER,
                                                SETTING_THUMBNAIL_MODE,
                                                id));
  if (!!cached_data)
    return decode_value(cached_data, MediaCodec());
  return ({});
}

static Medium
get_aphoto_medium(RequestID id)
{
  if (!id->misc->aphoto_medium) {
    id->misc->aphoto_medium
      = create_medium(get_aphoto_basename(id), aphoto_stat_file(id), id);
  }
  return id->misc->aphoto_medium;
}

static array(string)
get_aphoto_media_names(RequestID id)
{
  if (!id->misc->aphoto_media_names) {
    array(string) media_names = indices(get_aphoto_dir_stat(id));
    id->misc->aphoto_media_names
      = filter(media_names, lambda(string fn) {return is_medium_name(fn, id);});
  }
  return id->misc->aphoto_media_names;
}

static string
get_aphoto_mountpoint(RequestID id)
{
  if (!id->misc->aphoto_mountpoint) {
    string path = VFS.find_above(id->not_query, "", id);
    if (has_prefix(path, "//"))
      path = path[1..];
    string mountpoint = "/";
    foreach (id->conf->location_modules(), [string location,function callback])
      if (has_prefix(path, location)) {
        id->misc->aphoto_mountpoint = location;
        break;
      }
  }
  return id->misc->aphoto_mountpoint || "/";
}

static string
get_aphoto_mountpoint_relative_folder(RequestID id)
{
  if (!id->misc->aphoto_mountpoint_relative_folder) {
    string folder;
    sscanf(get_aphoto_file(id), get_aphoto_mountpoint(id) + "%s", folder);
    id->misc->aphoto_mountpoint_relative_folder = folder[..sizeof(folder) - 2];
  }
  return id->misc->aphoto_mountpoint_relative_folder;
}

static string
get_aphoto_next_browsing_page(RequestID id)
{
  if (!id->misc->aphoto_next_browsing_page)
    set_aphoto_adjacent_browsing_pages(id);
  return id->misc->aphoto_next_browsing_page;
}

static string
get_aphoto_previous_browsing_page(RequestID id)
{
  if (!id->misc->aphoto_previous_browsing_page)
    set_aphoto_adjacent_browsing_pages(id);
  return id->misc->aphoto_previous_browsing_page;
}

static string
get_aphoto_query(RequestID id)
{
  if (!id->misc->aphoto_query)
    if (id->query)
      id->misc->aphoto_query = "?" + id->query;
    else id->misc->aphoto_query = "";
  return id->misc->aphoto_query;
}

static array(int)
get_aphoto_raw_image_size(RequestID id)
{
  return Dims.dims()->get(open(get_aphoto_real_file(id), "r")) || ({0,0});
}

static string
get_aphoto_real_file(RequestID id)
{
  if (!id->misc->aphoto_real_file)
    id->misc->aphoto_real_file = id->conf->real_file(get_aphoto_file(id), id);
  return id->misc->aphoto_real_file;
}

static string
get_aphoto_shortened_relative_path(RequestID id)
{
  return shorten_path(get_aphoto_mountpoint_relative_folder(id),
                      get_aphoto_maximum_path_segments(id));
}

static string
get_aphoto_ssl_server(RequestID id)
{
  if (!id->misc->aphoto_ssl_server)
    id->misc->aphoto_ssl_server = query(SSL_CAPABLE_SERVER);
  return id->misc->aphoto_ssl_server;
}

static string
get_aphoto_suffix(RequestID id)
{
  if (!id->misc->aphoto_suffix)
    id->misc->aphoto_suffix = query(SETTING_SUFFIX);
  return id->misc->aphoto_suffix;
}

static string
get_aphoto_this_folder_page(RequestID id)
{
  if (!id->misc->aphoto_this_folder_page) {
    id->misc->aphoto_this_folder_page = "./" + get_aphoto_basename(id)
      + get_aphoto_folder_suffix(id);
  }
  return id->misc->aphoto_this_folder_page;
}

static string
get_aphoto_this_browsing_page(RequestID id)
{
  if (!id->misc->aphoto_this_browsing_page) {
    id->misc->aphoto_this_browsing_page = "./" + get_aphoto_basename(id)
      + get_aphoto_suffix(id);
  }
  return id->misc->aphoto_this_browsing_page;
}

static int(0..)
get_aphoto_thumbnail_size(RequestID id)
{
  if (!id->misc->aphoto_thumbnail_size)
    id->misc->aphoto_thumbnail_size = query(SETTING_THUMBNAIL_SIZE);
  return id->misc->aphoto_thumbnail_size;
}

static string
get_aphoto_tmpfile(RequestID id)
{
  if (!id->misc->aphoto_tmpfile) {
    string tmp_dir = dirname(get_aphoto_real_file(id));
    string tmpfile = combine_path(tmp_dir, get_aphoto_basename(id)) + ".tmp";
    int i = 0;
    while (!!file_stat(tmpfile)) {
      tmpfile = combine_path(tmp_dir, get_aphoto_basename(id)) + ".tmp." + i++;
    }
    id->misc->aphoto_tmpfile = tmpfile;
  }
  return id->misc->aphoto_tmpfile;
}

static int
get_aphoto_umask(RequestID id)
{
  if (!id->misc->aphoto_umask) {
    int umask;
    sscanf(query(SETTING_UMASK), "%o", umask);
    id->misc->aphoto_umask = umask;
  }
  return id->misc->aphoto_umask;
}

static string
get_aphoto_virtdir(RequestID id)
{
  if (!id->misc->aphoto_virtdir) {
    string dir = dirname(get_aphoto_file(id));
    id->misc->aphoto_virtdir = dir + ((dir[-1] != '/') ? "/" : "");
  }
  return id->misc->aphoto_virtdir;
}

static string
get_browsing_page(RequestID id)
{
  switch (query(SETTING_BROWSING_PAGE_ORIGIN)) {
  case PAGE_CONFIG_UI:
    return query(SETTING_BROWSING_PAGE);
  case PAGE_FROM_URL:
    return id->conf->try_get_file(query(SETTING_BROWSING_PAGE_URL), id)
      || "Error: aphoto could not fetch " + query(SETTING_BROWSING_PAGE_URL);
  case PAGE_BUILT_IN:
    return DEFAULT_BROWSING_PAGE;
  default:
    return "Erroneous browsing page origin!";
  }
}

static string
get_folder_page(RequestID id)
{
  switch (query(SETTING_FOLDER_PAGE_ORIGIN)) {
  case PAGE_CONFIG_UI:
    return query(SETTING_FOLDER_PAGE);
  case PAGE_FROM_URL:
    return id->conf->try_get_file(query(SETTING_FOLDER_PAGE_URL), id)
      || "Error: aphoto could not fetch " + query(SETTING_FOLDER_PAGE_URL);
  case PAGE_BUILT_IN:
    return DEFAULT_FOLDER_PAGE;
  default:
    return "Erroneous folder page location!";
  }
}

static array(mixed)
  get_image_data(string raw_image)
{
  if (is_jpeg_data(raw_image)) {
    mapping header = extract_jpeg_header(raw_image);
    if (header && header->comment) {
      return extract_aphoto_image_xml(header->comment);
    }
  }
  return ({0, 0, 0, 0, 0, 0, 0});
}

static array(int)
get_image_dims(string real_filename)
{
  Stdio.File fd = open(real_filename, "r");
  [int(0..) x, int(0..) y] = Dims.dims()->get(fd) || ({0,0});
  fd->close();
  return ({x, y});
}

static array(int)
  get_new_image_position_for_cache(int old_rot,
                                   int old_flip_h,
                                   int old_flip_v,
                                   int transform)
{
#if constant(Image.JPEG.ROT_90)
  int rot = (search(({Image.JPEG.ROT_90,Image.JPEG.ROT_180,Image.JPEG.ROT_270}),
                    transform) != -1) ? (transform - Image.JPEG.ROT_90 + 1) : 0;
  rot = (old_rot + rot) % 4;
  int flip_h = (old_flip_h + (transform == Image.JPEG.FLIP_H)) % 2;
  int flip_v = (old_flip_v + (transform == Image.JPEG.FLIP_V)) % 2;
  return ({rot, flip_h, flip_v});
#else
  return ({0, 0, 0});
#endif
}

static function(Medium,Medium:int(0..1))
get_sort_fun(mixed setting)
{
  switch (query(setting)) {
  case SORT_CRITERIUM_NAME:
    return sort_by_name;
  case SORT_CRITERIUM_NAME_REV:
    return reverse_sort_fun(sort_by_name);
  case SORT_CRITERIUM_MTIME:
    return sort_by_mtime;
  case SORT_CRITERIUM_MTIME_REV:
    return reverse_sort_fun(sort_by_mtime);
  }
}

static array(Medium)
get_sorted_aphoto_media(RequestID id)
{
  if (in_edit_mode(id))
    return get_sorted_aphoto_media_all(id);
  return filter(get_sorted_aphoto_media_all(id), is_visible);
}

static array(Medium)
get_sorted_aphoto_media_all(RequestID id)
{
  if (!id->misc->aphoto_sorted_media) {
    mapping(string:Stdio.Stat) stats = get_aphoto_dir_stat(id);
    if (!dir_changed_since_last_time(stats, id))
      return (id->misc->aphoto_media = get_aphoto_media_cache(id));
    array(Medium) media = ({});
    foreach (indices(stats), string filename) {
      if (is_medium_name(filename, id)) {
        media += ({create_medium(filename, stats[filename], id)});
      }
    }
    media = Array.sort_array(media, get_sort_fun(SETTING_MEDIA_SORT_ORDER));
    set_aphoto_media_cache(media, id);
    id->misc->aphoto_sorted_media = media;
  }
  return id->misc->aphoto_sorted_media;
}

static array(int)
get_thumbnail_size_from_file(string title, string filename, RequestID id)
{
  int(0..) x, y;
  if (should_make_title_field_on_thumbnail(title, filename, id)) {
    int(0..) side = get_aphoto_thumbnail_size(id)
      + IMAGE_THUMBNAIL_IMAGE_OVERHEAD;
    return ({side, side});
  }
  [x, y] = get_image_dims(id->conf->real_file(get_virtfile(filename, id), id));
  return
    complete_thumbnail_size(@shrink_size(x, y, get_aphoto_thumbnail_size(id)));
}

static array(int)
get_thumbnail_size_from_image(string title,
			      string filename,
			      Image.Image img,
			      RequestID id)
{
  if (should_make_title_field_on_thumbnail(title, filename, id))
    return ({get_aphoto_thumbnail_size(id), get_aphoto_thumbnail_size(id)});
  return ({img->xsize(), img->ysize()});
}

static string
get_virtfile(string filename, RequestID id)
{
  return get_aphoto_virtdir(id) + "/" + filename;
}

static mapping
handle_browsing_page_editing_post(RequestID id)
{
  if (!get_aphoto_editor(id)->authenticated())
    return id->conf->authenticate_throw(id, DATA_EDITING_PERMISSION);
  string raw_image = id->conf->try_get_file(get_aphoto_file(id), id);
  if (is_jpeg_data(raw_image) && supports_jpeg_transforms()) {
    int transform = get_aphoto_image_transform(id);
    string comment = make_image_comment(raw_image, transform, id);
    string new_raw_image = Image.JPEG.encode(raw_image,
                                             (["comment":comment,
                                               "transform":transform]));
    if (!get_aphoto_editor(id)->make_backup_file(get_aphoto_real_file(id),
                                                 get_aphoto_tmpfile(id))
        || !get_aphoto_editor(id)->write_file(get_aphoto_real_file(id),
                                              new_raw_image))
      return error_answer("Error when editing file! Check permissions on file "
                          "and hosting directory (for temporary file).", id);
    if (!saved_image_ok(raw_image, new_raw_image, id))
      return error_answer("An error occured when editing the file! "
                          + "Your old file resides at "
                          + basename(get_aphoto_tmpfile(id)), id);
    get_aphoto_editor(id)->remove_backup_file(get_aphoto_tmpfile(id));
  }
  return browsing_page_posted_redirect(id);
}

static mapping
handle_folder_editing_post(RequestID id)
{
  string real_file = combine_path(get_aphoto_real_file(id), DIR_INFO_FILENAME);
  if (!get_aphoto_editor(id)->authenticated())
    return id->conf->authenticate_throw(id, DATA_EDITING_PERMISSION);
  string title = id->variables->new_title;
  if (title == get_aphoto_shortened_relative_path(id))
    title = "";
  if (!get_aphoto_editor(id)
      ->write_file(real_file, make_dir_info(title,
                                            id->variables->new_long_desc,
                                            unixtime(id->variables->new_date))))
    return error_answer("Error when writing directory info file! Check "
                        "permissions on file and hosting directory", id);
  return Roxen.http_redirect(get_aphoto_this_folder_page(id), id);
}

static mapping
handle_request(RequestID id)
{
  NOCACHE();
  mixed error = catch {
    if (is_data(id))
      return generate_browsing_page(id);
    if (is_folder(id) && index_file_present(id)
        || !aphoto_has_folder_suffices(id))
      return Roxen.http_redirect("./" + get_aphoto_query(id), id);
    else
      return generate_folder_page(id);
  };
  werror("%O\n", error);
  return 0;
}

static int(0..1)
in_edit_mode(RequestID id)
{
  return !!id->variables->edit && id->variables->edit == "true";
}

static int(0..1)
in_memento_mode(RequestID id)
{
  if (!id->misc->aphoto_in_memento_mode)
    id->misc->aphoto_in_memento_mode = low_query_in_memento_mode();
  return id->misc->aphoto_in_memento_mode;
}

static int(0..1)
index_file_present(RequestID id)
{
  RoxenModule dirs;
  if (dirs = id->conf->find_module("directories")) {
    foreach (dirs->query("indexfiles"), string filename) {
      array s;
      string file = get_aphoto_virtdir(id) + filename;
      if ((s = id->conf->stat_file(file, id)) && (s[ST_SIZE] >= 0))
        return 1;
    }
  }
  return 0;
}

static string
initial_user_auth_mode() {
  if (running_windows())
    return USER_AUTH_MODE_CONFIGIFC;
  return USER_AUTH_MODE_USERDB;
}

static int(0..1)
insecure(RequestID id) {
  NOCACHE();
  return id->port_obj->name != "https";
}

static int(0..1)
//! To avoid "up" arrows on top pages. Also handles ~.
is_at_root_mountpoint(RequestID id)
{
  string mountpoint = get_aphoto_mountpoint(id) || "";
  string endchar = mountpoint[strlen(mountpoint) - 1..strlen(mountpoint) - 1];
  string mountpoint_rel = get_aphoto_mountpoint_relative_folder(id);
  if (is_folder(id) && mountpoint_rel != "")
    mountpoint_rel += "/";
  string query = endchar + mountpoint_rel;
  return String.count(query, "/") < 2;
}

static int(0..1)
is_data(RequestID id)
{
  return is_file(aphoto_stat_file(id));
}

static int(0..1)
is_dir(Stdio.Stat stat)
{
  return ((array(int)) stat)[ST_SIZE] == -2;
}

static int(0..1)
is_editable_data(RequestID id)
{
  return is_data(id) && is_jpeg_name(get_aphoto_basename(id), id);
}

static int(0..1)
is_file(Stdio.Stat stat)
{
  return ((array(int)) stat)[ST_SIZE] >= 0;
}

static int(0..1)
is_folder(RequestID id)
{
  return is_dir(aphoto_stat_file(id));
}

static int(0..1)
is_horizontal(Image.Image image)
{
  if (image->xsize() > image->ysize()) {
    return 1;
  }
  return 0;
}

static int(0..1)
is_image_name(string filename, RequestID id)
{
  return has_prefix(id->conf->type_from_filename(filename), "image/");
}

static int(0..1)
is_jpeg_name(string filename, RequestID id)
{
  return has_prefix(id->conf->type_from_filename(filename), "image/jpeg");
}

static int(0..1)
is_movie_name(string filename, RequestID id)
{
  return has_prefix(id->conf->type_from_filename(filename), "video/");
}

static int(0..1)
is_medium_name(string filename, RequestID id)
{
  return is_image_name(filename, id);// || is_movie_name(filename, id);
}

static int(0..1)
is_my_request(RequestID id)
{
  return has_suffix(id->not_query, get_aphoto_suffix(id));
}

static int(0..1)
is_jpeg_data(string raw_data)
{
  return !(sizeof(raw_data) < 2) && raw_data[0..1] == "\377\330";
}

static int(0..1)
is_visible(Medium medium)
{
  return !medium->hidden;
}

int(0..1)
low_query_in_memento_mode()
{
  return query(SETTING_THUMBNAIL_MODE) == THUMBNAIL_MODE_MEMENTO;
}

static Image.image
make_border(Image.image in, int border_width)
{
  Image.image border = Image.image(in->xsize() + border_width * 2,
				   in->ysize() + border_width * 2,
				   Image.Color.white);
  border->paste(in, border_width, border_width);
  return border;
}

static string
make_dir_info(string title, string long_desc, int date)
{
  return make_xml(APHOTO,
		  "\n" + make_xml(TAG_DIR_DESC_SHORT, title) + "\n" +
		  make_xml(TAG_DIR_DESC_LONG, long_desc) + "\n" +
		  make_xml(TAG_DIR_DATE, (string) date) + "\n");
}

static string
make_image_comment(string old_raw_image, int transform, RequestID id)
{
  mapping header = extract_jpeg_header(old_raw_image);
  string comment = "";
  if (header)
    comment = header->comment || "";
  string unused;
  int old_rot, old_fliph, old_flipv, hidden, unused2;
  if (comment && comment != "") {
    [unused, unused, unused2, hidden, old_rot, old_fliph, old_flipv]
      = extract_aphoto_image_xml(comment);
    string old_data = make_xml(TAG_APHOTO, extract_xml(TAG_APHOTO, comment));
    //tidy up old comments (if many)
    while (comment - old_data != comment) {
      comment -= old_data;
      old_data = make_xml(TAG_APHOTO, extract_xml(TAG_APHOTO, comment));
    }
  }
  [int rot, int fliph, int flipv] = get_new_image_position_for_cache(old_rot,
                                                                     old_fliph,
                                                                     old_flipv,
                                                                     transform);
  string new_comment
    = comment + make_image_info(id->variables->new_title,
                                id->variables->new_long_desc,
                                unixtime(id->variables->new_date),
                                hidden && !id->variables->unhide
                                || !!id->variables->hide_and_next,
                                rot,
                                fliph,
                                flipv);
  return new_comment;
}

static string
make_image_info(string title,
                string long_desc,
                int date,
                int hidden,
                int(0..3) rot,
                int(0..1) flip_h,
                int(0..1) flip_v)
{
  return make_xml(APHOTO,
		  "\n" + make_xml(TAG_IMAGE_DESC_SHORT, title) + "\n" +
		  make_xml(TAG_IMAGE_DESC_LONG, long_desc) + "\n" +
		  make_xml(TAG_IMAGE_DATE, (string) date) + "\n" +
                  (hidden ? make_xml(TAG_HIDDEN, 0) + "\n" : "") +
                  make_xml(TAG_IMAGE_POSITION,
                           "" + rot + "-" + flip_h + "-" + flip_v));
}

static Image.Image
make_resized_image(string raw_image, int(0..) wanted_maxsize)
{
  Image.Image image = Image._decode(raw_image)->img;
  int(0..) x_size = wanted_maxsize;
  int(0..) y_size = x_size;

  if (is_horizontal(image))
    y_size = 0;
  else
    x_size = 0;

  if (image->xsize() < x_size) x_size=0;
  if (image->ysize() < y_size) y_size=0;
  if (x_size||y_size) image = image->scale(x_size, y_size);
  return image;
}

static string
make_xml(string tag, string text_block)
{
  if (text_block != 0) {
    return "<" + tag + ">" + text_block + "</" + tag + ">";
  }
  else {
    return "<" + tag + "/>";
  }
}

static int(0..1)
memento_mode_invisibility_callback(RequestID id, Variable.Variable var)
{
  return !low_query_in_memento_mode();
}

static mapping
redirect_to_secure_server(RequestID id)
{
  return Roxen.http_redirect(query(SSL_CAPABLE_SERVER)
			     + id->not_query
			     + get_aphoto_query(id));
}

static function(Medium,Medium:int(0..1))
reverse_sort_fun(function(Medium,Medium:int(0..1)) sort_fun)
{
  return lambda(Medium left, Medium right) {return sort_fun(right, left);};
}

static int(0..1)
running_windows() {
#ifdef __NT__
  return 1;
#else
  return 0;
#endif
}

static int(0..1)
running_locally(RequestID id) {
  return "127.0.0.1" == id->remoteaddr;
}

static int(0..1)
saved_image_ok(string old_image, string new_image, RequestID id)
{
  string saved_image = id->conf->try_get_file(get_aphoto_file(id), id);
  if (saved_image != new_image)
    return 0;
  if (!dimensions_equal_or_rotated(old_image, new_image))
    return 0;
  return 1;
}

static void
set_aphoto_adjacent_browsing_pages(RequestID id)
{
  array(Medium) media = get_sorted_aphoto_media(id);
  for (int i = 0; i < sizeof(media); ++i) {
    if (media[i]->name == get_aphoto_basename(id)) {
      if (i != 0)
        id->misc->aphoto_previous_browsing_page =
          "./" + media[i - 1]->name + get_aphoto_suffix(id);
      if ((i + 1) != sizeof(media))
        id->misc->aphoto_next_browsing_page =
          "./" + media[i + 1]->name + get_aphoto_suffix(id);
      break;
    }
  }
}

static void
set_aphoto_dir_stat_cache(mapping(string:Stdio.Stat) stats, RequestID id)
{
  string key = get_aphoto_cache_key(DIR_STAT_CACHE,
				    SETTING_MEDIA_SORT_ORDER,
				    SETTING_THUMBNAIL_MODE,
				    id);
  string value = encode_value(stats_to_array(stats));
  cache.set_session_data(value, key, CACHE_TIME, 1);
}

static void
set_aphoto_media_cache(array(Medium) media, RequestID id)
{
  cache.set_session_data(encode_value(media, MediaCodec()),
                         get_aphoto_cache_key(MEDIA_CACHE,
                                              SETTING_MEDIA_SORT_ORDER,
                                              SETTING_THUMBNAIL_MODE,
                                              id),
                         CACHE_TIME,
                         1);
}

static void
set_global_form_variables(Medium medium, RequestID id)
{
  id->variables->debug_info = "<aphoto-debug/>";
}

static array(string)
split_title_in_lines(string title, int(0..1) hidden)
{
  return map(sprintf("%-="
                     + query(SETTING_LETTERS_PER_LINE)
                     + "s",
                     hidden ? "Hidden image!" : title)/"\n",
             String.trim_whites);
}

static string
shorten_path(string path, int max_parts)
{
  array(string) parts = path / "/";
  if (max_parts >= sizeof(parts - ({""})))
    return path;
  string new_path
    = combine_path("/", parts[enumerate(max_parts, 1, -max_parts)[*]] * "/");
  if (new_path != path)
    return "..." + new_path;
  return path;
}

static array(int)
shrink_size(int x, int y, int size)
{
  if (x <= size && y <= size)
    return ({x, y});
  float scale = size / (float) max(x, y);
  return ({(int) (x * scale), (int) (y * scale)});
}

static int(0..1)
should_change_umask(RequestID id) {
  if (!id->misc->aphoto_should_change_umask) {
    id->misc->aphoto_should_change_umask = query(SETTING_SET_UMASK);
  }
  return id->misc->aphoto_should_change_umask;
}

static int(0..1)
should_make_scaled_backdropped_images(RequestID id)
{
  if (!id->misc->aphoto_scaled_backdrops)
    id->misc->aphoto_scaled_backdrops =
      query(SETTING_BROWSING_IMAGE_MODE) == BROWSING_IMAGE_MODE_SCALED_SHADOWED;
  return id->misc->aphoto_scaled_backdrops;
}

static int(0..1)
should_make_title_field_on_thumbnail(string title, string filename, RequestID i)
{
  return in_memento_mode(i) && title && title != "" && title != filename;
}

static int(0..1)
should_redirect_to_ssl(RequestID id) {
  if (!id->misc->aphoto_should_redirect_ssl) {
    id->misc->aphoto_should_redirect_ssl = query(SETTING_REDIRECT_SECURE);
  }
  return id->misc->aphoto_should_redirect_ssl;
}

static int(0..1)
sort_by_mtime(Medium left, Medium right) {
  return left->unix_time > right->unix_time;
}

static int(0..1)
sort_by_name(Medium left, Medium right) {
  return left->name > right->name;
}

static mapping(string:array(int))
stats_to_array(mapping(string:Stdio.Stat) stats)
{
  return (mapping(string:array(int))) stats;
}

static string
store_thumbnail(Medium medium, string dir, string int_path, RequestID id)
{
  string virt_path = dir + medium->name;
  return int_path + store_in_image_cache((["file":virt_path,
                                           "hidden":medium->hidden,
                                           "memento":in_memento_mode(id),
                                           "type":TYPE_THUMBNAIL,
                                           "title":medium->title,
                                           "rotation":medium->rotation,
                                           "flip_h":medium->flip_h,
                                           "flip_v":medium->flip_v,
                                           "format":"png",
					   "unix_time":medium->unix_time]), id);
}

static string
store_browsing_image(Medium medium, RequestID id)
{
  string int_path = get_aphoto_internal_location(id);
  return int_path + store_in_image_cache((["file":get_aphoto_file(id),
                                           "hidden":medium->hidden,
                                           "type":TYPE_BROWSING_IMAGE,
                                           "rotation":medium->rotation,
                                           "flip_h":medium->flip_h,
                                           "flip_v":medium->flip_v,
                                           "format":"jpeg",
					   "unix_time":medium->unix_time]), id);
}

static string
store_in_image_cache(mapping m, RequestID id)
  //! The @[store_in_image_cache()] method exists because of the
  //! fact that the @[image_cache->store()] adds an extra user
  //! attribute to the key mapping, making the logged in edit pages
  //! reload all thumbnails again(!). This is avoided by removing
  //! the authenticated_user temporarily.
{
  User pushed_user = id->misc->authenticated_user;
  id->misc -= (["authenticated_user":id->misc->authenticated_user]);
  string ret = image_cache->store(m, id);
  id->misc->authenticated_user = pushed_user;
  return ret;
}

static int(0..1)
supports_jpeg_transforms()
{
#if constant(Image.JPEG.ROT_90)
  return 1;
#else
  return 0;
#endif
}

static string
stringdate(int unix_time)
{
  mapping(string:int) tm = gmtime(unix_time);
  return sprintf("%04d-%02d-%02d", tm->year + 1900, tm->mon + 1, tm->mday);
}

static string
stringtime(int unix_time)
{
  mapping(string:int) tm = gmtime(unix_time);
  return sprintf("%02d:%02d:%02d", tm->hour, tm->min, tm->sec);
}

static string
strip_mountpoint(string query, RequestID id)
{
  return "not_implemented";
}

static int
unixtime(string timestring)
{
  // exif datetime is in YYYY:MM:DD HH:MM:SS format
  int yyyy, mon, dd, hh, mm, ss;
  if (sscanf(timestring, "%d:%d:%d %d:%d:%d", yyyy, mon, dd, hh, mm, ss) != 6 &&
      sscanf(timestring, "%d-%d-%d %d:%d:%d", yyyy, mon, dd, hh, mm, ss) != 6)
    return (int) timestring;
  if (mon == 0)
    return 0;
  if (yyyy <= 1900 && yyyy >= 100)
    return 0;
  if (yyyy < 100) {
    if (yyyy < 30) {
      yyyy += 2000;
    } else {
      yyyy += 1900;
    }
  }
  mapping(string:int) tm = ([
    "hour":hh,
    "mday":dd,
    "min":mm,
    "mon":mon - 1,
    "sec":ss,
    "year":yyyy - 1900,
    "isdst":0,
    "timezone":0
  ]);
  return mktime(tm);
}

/**
 * Tags.
 */
class TagAphotoArrow
{
  inherit RXML.Frame;

  int(0..1)
  is_active()
  {
    return !args || !args->active || args->active != "false";
  }

  string
  link(string active_image, string inactive_image, RequestID id)
  {
    string arrow = is_active() ? active_image : inactive_image;
    args->src = get_aphoto_internal_location(id) + arrow;
    mapping link_args = args;
    if (link_args->active)
      link_args -= (["active":args->active]);
    return Roxen.parse_rxml(Roxen.make_tag("imgb0", link_args, 1), id);
  }
}

class TagAphotoArrowLeft
{
  inherit RXML.Tag;
  constant name = "aphoto-arrow-left";
  constant flags = RXML.FLAG_EMPTY_ELEMENT;

  class Frame
  {
    inherit TagAphotoArrow;

    array do_return(RequestID id)
    {
      result = link(INTERNAL_ARROW_LEFT, INTERNAL_ARROW_LEFT_NONACTIVE, id);
      return 0;
    }
  }
}

class TagAphotoArrowUp
{
  inherit RXML.Tag;
  constant name = "aphoto-arrow-up";
  constant flags = RXML.FLAG_EMPTY_ELEMENT;

  class Frame
  {
    inherit TagAphotoArrow;

    array do_return(RequestID id)
    {
      result = link(INTERNAL_ARROW_UP, INTERNAL_ARROW_UP_NONACTIVE, id);
      return 0;
    }
  }
}

class TagAphotoArrowRight
{
  inherit RXML.Tag;
  constant name = "aphoto-arrow-right";
  constant flags = RXML.FLAG_EMPTY_ELEMENT;

  class Frame
  {
    inherit TagAphotoArrow;

    array do_return(RequestID id)
    {
      result = link(INTERNAL_ARROW_RIGHT, INTERNAL_ARROW_RIGHT_NONACTIVE, id);
      return 0;
    }
  }
}

class TagAphotoDebug
{
  inherit RXML.Tag;
  constant name = "aphoto-debug";
  constant flags = RXML.FLAG_EMPTY_ELEMENT;

  class Frame
  {
    inherit RXML.Frame;
    array do_return(RequestID id)
    {
      result = "<pre>Debug data:\n"
        + "\nAphotofile: " + get_aphoto_file(id)
        + "\nRealfile: " + get_aphoto_real_file(id)
        + "\nIs dir: " + is_folder(id)
        + "\nIs medium: " + is_data(id)
        + "\nIs at root: " + is_at_root_mountpoint(id)
        + "\nBasename: " + get_aphoto_basename(id)
        + "\nVirtdir: " + get_aphoto_virtdir(id)
        + "\nMountpoint: " + get_aphoto_mountpoint(id)
        + "\nMountpoint rel: " + get_aphoto_mountpoint_relative_folder(id)
        + "\n</pre>";
      return 0;
    }
  }
}

class TagAphotoForm
{
  inherit RXML.Tag;
  constant name = "aphoto-form";

  class Frame
  {
    static string
    create_edit_tag(mapping args, string content, RequestID id)
    {
      if (is_folder(id))
	args->action = get_aphoto_this_folder_page(id);
      else
	args->action = get_aphoto_this_browsing_page(id);
      args->action += EDIT_QUERY;
      return Roxen.make_tag("form", args, 0) + content + "</form>";
    }

    inherit RXML.Frame;
    array do_return(RequestID id)
    {
      result = create_edit_tag(args, content, id);
      return 0;
    }
  }
}

class TagAphotoLink
{
  inherit RXML.Tag;
  constant name = "aphoto-link";
  //  constant flags = RXML.FLAG_GET_EVALED_CONTENT;

  static string
  create_edit_tag(mapping args, string content, RequestID id)
  {
    if (is_folder(id))
      args->href += get_aphoto_this_folder_page(id);
    else
      args->href += get_aphoto_this_browsing_page(id);
    args->href += EDIT_QUERY;
    return make_anchor(args, content);
  }

  static string
  create_exit_tag(mapping args, string content, RequestID id)
  {
    if (is_folder(id))
      args->href += get_aphoto_this_folder_page(id);
    else
      args->href += get_aphoto_this_browsing_page(id);
    return make_anchor(args, content);
  }

  static string
  create_next_tag(mapping args, string content, RequestID id)
  {
    if (is_folder(id)) {
      return 0;
    }
    else {
      string next_page = get_aphoto_next_browsing_page(id);
      if (!next_page || next_page == "") return 0;
      args->href = next_page + get_aphoto_query(id);
    }
    return make_anchor(args, content);
  }

  static string
  create_parent_tag(mapping args, string content, RequestID id)
  {
    if (is_folder(id)) {
      if (is_at_root_mountpoint(id))
	return 0;
      args->href = "../" + get_aphoto_folder_suffix(id) + get_aphoto_query(id);
    }
    else {
      args->href = "./" + get_aphoto_folder_suffix(id)+ get_aphoto_query(id);
    }
    return make_anchor(args, content);
  }

  static string
  create_previous_tag(mapping args, string content, RequestID id)
  {
    if (is_folder(id)) {
      return 0;
    }
    else {
      string previous_page = get_aphoto_previous_browsing_page(id);
      if (!previous_page || previous_page == "") return 0;
      args->href = previous_page + get_aphoto_query(id);
    }
    return make_anchor(args, content);
  }

  static string
  create_medium_tag(mapping args, string content, RequestID id)
  {
    args->href += get_aphoto_basename(id);
    return make_anchor(args, content);
  }

  static string
  create_stylesheet_tag(mapping args, RequestID id)
  {
    args->href = get_aphoto_internal_location(id) + INTERNAL_CSS;
    args->rel = "stylesheet";
    args->title = "default aphoto style";
    args->type = "text/css";
    return Roxen.make_tag("link", args, 1);
  }

  static string
  make_anchor(mapping args, string content)
  {
    return Roxen.make_tag("a", args, 0) + content + "</a>";
  }

  class Frame
  {
    inherit RXML.Frame;
    array do_return(RequestID id) 
    {
      if (!args->href) {
	args->href = "./";
      }
      string type = args->type;
      args -= (["type":args->type]);
      switch (type) {
      case "edit":
	result = create_edit_tag(args, content, id);
	break;
      case "exit":
	result = create_exit_tag(args, content, id);
	break;
      case "medium":
	result = create_medium_tag(args, content, id);
	break;
      case "next":
	result = create_next_tag(args, content, id);
	break;
      case "parent":
	result = create_parent_tag(args, content, id);
	break;
      case "previous":
	result = create_previous_tag(args, content, id);
	break;
      case "stylesheet":
	result = create_stylesheet_tag(args, id);
	break;
      case 0:
	RXML.run_error("No link type supplied!\n");
	break;
      default:
	RXML.run_error("Unknown link type %s!\n", args->type);
	break;
      }
      if (!result) {
	RXML_CONTEXT->misc[" _ok"] = 0;
	result = "";
      } else RXML_CONTEXT->misc[" _ok"] = 1;
      return 0;
    }
  }
}

class TagBrowsingImage
{
  inherit RXML.Tag;
  constant name = "aphoto-browsing-image";
  constant flags = RXML.FLAG_EMPTY_ELEMENT;

  class Frame
  {
    inherit RXML.Frame;
    array do_return(RequestID id) {
      string width, height;
      string src;
      if (should_make_scaled_backdropped_images(id)) {
	width = get_aphoto_browsing_image_size(id)[0];
	height = get_aphoto_browsing_image_size(id)[1];
	src = store_browsing_image(get_aphoto_medium(id), id);
      }
      else {
	width = (string) get_aphoto_raw_image_size(id)[0];
	height = (string) get_aphoto_raw_image_size(id)[1];
	src = "./" + get_aphoto_basename(id);
      }
      args += (["src":src, "width":width, "height":height]);
      result = Roxen.parse_rxml(Roxen.make_tag("imgb0", args, 1), id);
      return 0;
    }
  }
}

class TagDate
{
  inherit RXML.Tag;
  constant name = "aphoto-date";
  constant flags = RXML.FLAG_EMPTY_ELEMENT;

  class Frame
  {
    inherit RXML.Frame;
    array do_return(RequestID id)
    {
      result = stringdate(get_aphoto_medium(id)->unix_time);
      return 0;
    }
  }
}

class TagLongDesc
{
  inherit RXML.Tag;
  constant name = "aphoto-description";
  constant flags = RXML.FLAG_EMPTY_ELEMENT;

  class Frame
  {
    inherit RXML.Frame;
    array do_return(RequestID id)
    {
      result = get_aphoto_medium(id)->long_desc || "";
      return 0;
    }
  }
}

class TagEmitFolders {
  inherit RXML.Tag;
  constant name = "emit";
  constant plugin_name = "folders";

  array get_dataset(mapping args, RequestID id)
  {
    array ret = ({});
    string dir = get_aphoto_virtdir(id);
    string image = get_aphoto_internal_location(id) + INTERNAL_FOLDER;
    string width = (string) Image._decode(APHOTO_FOLDER)->img->xsize();
    string height = (string) Image._decode(APHOTO_FOLDER)->img->ysize();
    string suffix = "/" + get_aphoto_folder_suffix(id) + get_aphoto_query(id);
    foreach(get_aphoto_folders(id), Medium medium) {
      ret += ({(["file":medium->name,
                 "link":medium->name + suffix,
                 "long_desc":medium->long_desc,
                 "image":image,
                 "width":width,
                 "height":height,
                 "title":medium->title])});
    }
    return ret;
  }
}

class TagGalleryName
{
  inherit RXML.Tag;
  constant name = "aphoto-gallery";
  constant flags = RXML.FLAG_EMPTY_ELEMENT;

  class Frame
  {
    inherit RXML.Frame;
    array do_return(RequestID id)
    {
      result = get_aphoto_gallery_name(id);
      return 0;
    }
  }
}

class TagIfHidden {
  inherit RXML.Tag;
  constant name = "if";
  constant plugin_name = "hidden";

  int eval(string dbg, RequestID id, mapping m) {
    return !!get_aphoto_medium(id)->hidden;
  }
}

class TagImgBorderZeroIfNetscape
{
  inherit RXML.Tag;
  constant name = "imgb0";
  constant flags = RXML.FLAG_EMPTY_ELEMENT;

  class Frame
  {
    inherit RXML.Frame;
    array do_return(RequestID id)
    {
      array client = id->client;
      if (client && sizeof(client) && client[0][0..8] == "Mozilla/4") {
	args->border = 0;
      }
      result = Roxen.make_tag("img", args, 1);
      return 0;
    }
  }
}

class TagTime
{
  inherit RXML.Tag;
  constant name = "aphoto-time";
  constant flags = RXML.FLAG_EMPTY_ELEMENT;

  class Frame
  {
    inherit RXML.Frame;
    array do_return(RequestID id)
    {
      result = stringtime(get_aphoto_medium(id)->unix_time);
      return 0;
    }
  }
}

class TagTitle
{
  inherit RXML.Tag;
  constant name = "aphoto-title";
  constant flags = RXML.FLAG_EMPTY_ELEMENT;

  class Frame
  {
    inherit RXML.Frame;
    array do_return(RequestID id)
    {
      string title = get_aphoto_medium(id)->title || "";
      if (title == "" && !!args && !!args->fallback && args->fallback == "path")
	result = get_aphoto_shortened_relative_path(id);
      else
	result = title;
      if (!!args && !!args->prepend && !!result && result != "")
	result = args->prepend + result;
      return 0;
    }
  }
}

class TagEmitThumbnails {
  inherit RXML.Tag;
  constant name = "emit";
  constant plugin_name = "thumbnails";

  array get_dataset(mapping args, RequestID id)
  {
    array ret = ({});
    string dir = get_aphoto_virtdir(id);
    string internal_path = get_aphoto_internal_location(id);
    string suffix = get_aphoto_suffix(id) + get_aphoto_query(id);
    foreach(get_sorted_aphoto_media(id), Medium medium) {
      string thumbnail = store_thumbnail(medium, dir, internal_path, id);
      string popup = in_memento_mode(id) ? medium->long_desc : medium->title;
      array(string) popup_desc_lines = (popup||"")/"\n";
      if (sizeof(popup_desc_lines) > 1)
        popup_desc_lines[0] = popup_desc_lines[0] - "\n" - "\r" + " <...>";
      ret += ({(["link":"./" + medium->name + suffix,
                 "thumbnail":thumbnail,
                 "long_desc":medium->long_desc,
                 "popup_desc":popup_desc_lines[0],
                 "width":medium->thumb_w,
                 "height":medium->thumb_h,
                 "title":medium->title])});
    }
    return ret;
  }
}

/**
 * Internal classes.
 */
static class Codecable() {
  mapping _encode() {
    mapping m = ([]);
    foreach(indices(this), string i) {
      mixed v = this[i];
      if(intp(v) || stringp(v) || floatp(v))
        m[i] = v;
    }
    return m;
  }

  void _decode(mapping m) {
    foreach(m; string i; mixed v)
      this[i] = v;
  }
}

static class MediaCodec {
  static constant JPEG_IMAGE_MEDIUM_CLASS_ID = "JpegImageMedium";
  static constant IMAGE_MEDIUM_CLASS_ID = "ImageMedium";
  static constant MOVIE_MEDIUM_CLASS_ID = "MovieMedium";

  //Called by encode_value for each function, program and object.
  mixed nameof(mixed x) {
    if (x == JpegImageMedium) return JPEG_IMAGE_MEDIUM_CLASS_ID;
    if (x == ImageMedium) return IMAGE_MEDIUM_CLASS_ID;
    if (x == MovieMedium) return MOVIE_MEDIUM_CLASS_ID;
    return ([])[0];
  }

  // Called by decode_value to do the reverse of nameof() for functions.
  function functionof(mixed s)
  {
    switch (s) {
    case JPEG_IMAGE_MEDIUM_CLASS_ID:
      return JpegImageMedium;
    case IMAGE_MEDIUM_CLASS_ID:
      return ImageMedium;
    case MOVIE_MEDIUM_CLASS_ID:
      return MovieMedium;
    default:
      error ("Don't know how to decode function %O.\n", s);
    }
  }

  // Called by decode_value to do the reverse of nameof() for programs.
  program programof(mixed s)
  {
    switch (s) {
    case JPEG_IMAGE_MEDIUM_CLASS_ID:
      return JpegImageMedium;
    case IMAGE_MEDIUM_CLASS_ID:
      return ImageMedium;
    case MOVIE_MEDIUM_CLASS_ID:
      return MovieMedium;
    default:
      error("Don't know how to decode program %O.\n", s);
    }
  }

  // Called by decode_value to do the reverse of nameof() for objects.
  object objectof(mixed s)
  {
    error("Don't know how to decode object %O.\n", s);
  }

  // Called by encode_value to encode the contents of an object.
  mixed encode_object(object o)
  {
    return o->_encode();
  }

  // Called by decode_value to fill in the contents of an object.
  void decode_object (object o, mixed x)
  {
    o->_decode (x);
  }
}

static class Medium {
  string name;
  string title;
  string long_desc;
  int unix_time;
  int(0..1) hidden;
  int thumb_w;
  int thumb_h;
  int(0..3) rotation;
  int(0..1) flip_h;
  int(0..1) flip_v;
}

static class AphotoDir {
  inherit Medium;
  inherit Codecable;

  void create(string dirname, Stdio.Stat stat, RequestID id) {
    name = dirname;
    string virt_file;
    if (dirname && dirname != "")
      virt_file = get_virtfile(dirname, id) + "/" + DIR_INFO_FILENAME;
    else
      virt_file = get_virtfile(".aphoto_dir_info.xml", id);
    if (virt_file) {
      string xml = id->conf->try_get_file(virt_file, id);
      [title, long_desc, unix_time, hidden] = extract_aphoto_folder_xml(xml);
    }
    if (!title || title == "")
      title = dirname;
    unix_time = unix_time || stat->mtime;
  }
}

static class ImageMedium {
  inherit Medium;
  inherit Codecable;

  void create(string filename, Stdio.Stat stat, RequestID id) {
    name = filename;
    title = filename;
    long_desc = "";
    unix_time = stat->mtime;
    [thumb_w, thumb_h] = get_thumbnail_size_from_file(title, filename, id);
  }
}

static class JpegImageMedium {
  inherit Medium;
  inherit Codecable;

  void create(string filename, Stdio.Stat stat, RequestID id) {
    name = filename;
    string virt_file = get_virtfile(filename, id);
    string raw_image = id->conf->try_get_file(virt_file, id);
    [title, long_desc, unix_time, hidden, rotation, flip_h, flip_v]
      = get_image_data(raw_image);
    unix_time = unix_time
      || exif_time(id->conf->real_file(virt_file, id))
      || stat->mtime;
    title = title || filename;
    [thumb_w, thumb_h] = get_thumbnail_size_from_file(title, filename, id);
  }
}

static class MovieMedium {
  inherit Medium;
  inherit Codecable;

  void create(string filename, Stdio.Stat stat, RequestID id) {
  }
}

static class AlbumEditor {
  RequestID id;
  int umask;

  int(0..1) authenticated() {
    return !in_edit_mode(get_id()) && !editing_data_posted(get_id());
  }

  void create(RequestID new_id) {
    id = new_id;
  }

  RequestID get_id() {
    return id;
  }

  int(0..1) make_backup_file(string file, string tmp_file) {
    push_umask();
    int ret;
    catch {
      ret = Stdio.cp(file, tmp_file);
    };
    pop_umask();
    return ret;
  }

  void push_umask() {
    if (should_change_umask(get_id()))
        umask = System.umask(get_aphoto_umask(get_id()));
  }

  void pop_umask() {
    if (should_change_umask(get_id()))
        System.umask(umask);
  }

  int(0..1) remove_backup_file(string file) {
    return rm(file);
  }

  int(0..1) write_file(string file, string content) {
    push_umask();
    int ret;
    catch {
      ret = Stdio.write_file(file, content);
    };
    pop_umask();
    return ret;
  }
}

static class UserDBEditor {
  inherit AlbumEditor;

  int(0..1) authenticated() {
    if (!::authenticated())
      return !!get_id()->conf->authenticate(get_id());
    return 1;
  }

  Privs get_privs(string cause) {
    User user = get_id()->conf->authenticate(get_id());
    return Privs(cause, user->uid(), user->gid());
  }

  int(0..1) make_backup_file(string file, string tmp_file) {
    User user = get_id()->conf->authenticate(get_id());
    Privs p = get_privs("Creating backup file for photo album as user");
    int ret = ::make_backup_file(file, tmp_file);
    p = 0;
    return ret;
  }

  int(0..1) remove_backup_file(string file) {
    Privs p = get_privs("Removing file " + file + " as user.");
    int ret = rm(file);
    p = 0;
    return ret;
  }

  int(0..1) write_file(string file, string content) {
    Privs p = get_privs("Writing file " + file + " as user.");
    System.umask();
    int ret = ::write_file(file, content);
    p = 0;
    return ret;
  }
}

static class ConfigUIEditor {
  inherit AlbumEditor;

  int(0..1) authenticated() {
    if (!::authenticated())
      return
	!!roxen.configuration_authenticate(get_id(), DATA_EDITING_PERMISSION);
    return 1;
  }
}


/**
 * Images.
 */
static constant APHOTO_FOLDER = "GIF87a \0 \0\0\0\0\0\200\0\0\0\200\0\200\200\0\0\0\200\200\0\200\0\200\200\231f3\231f3\0\231\231\231\231\231f\231""3\231\0fff\231fff3f\0333\2313f333\0\0\0\231\0f\0""3\231ff3\0\231f3\0\231\231\231\231\231f\231""3\231\0fff\231fff3f\0333\2313f333\0\0\0\0\231\0f\0""3\0\0\231\231\231\231\231f\2313\231\0\231\231\231\231\231f\2313\231\0\231\231\231\231\231\231\231\231\231f\231\231""3\231\231\0\231f\231f\231f\231\231ff\231f3\231f\0\231""3\231""3\231""3\231\231""3f\231""33\231""3\0\231\0\231\0\231\0\231\231\0f\231\0""3\231\0\0fff\231fff3f\0fff\231fff3f\0f\231f\231f\231\231f\231ff\231""3f\231\0ffffff\231fffff3ff\0f3f3f3\231f3ff33f3\0f\0f\0f\0\231f\0ff\0""3f\0\0""333\231""3f333\0""333\231""3f333\0""3\2313\2313\231\231""3\231f3\231""33\231\0""3f3f3f\231""3ff3f33f\0""333333\231""33f33333\0""3\03\03\0\231""3\0f3\0""33\0\0\0\0\231\0f\03\0\0\0\231\0f\03\0\0\0\231\0\231\0\231\231\0\231f\0\231""3\0\231\0\0f\0f\0f\231\0ff\0f\0\0f3\0""3\0""3\0""3\231\0""3f\0""33\0""3\0\0\0\0\0\231\0\0f\0\0""3\0\0\0\0\0\0\210\0\0w\0\0U\0\0D\0\0\"\0\0\0\0\0\0\0\0\0\210\0\0w\0\0U\0\0D\0\0\"\0UUUwwwwwwDDD\"\"\"\21\21\21\0\0w\0\0U\0\0D\0\0\"𠠤\200\200\200\0\0\0\0\0\0\0\0\0!\4\1\0\0\1\0,\0\0\0\0 \0 \0\0\b\0\3\b\34H\203\b\23*\\Ȱ\207\20#J\234h\20\200\213\30)\6\0Х\217\0&\2GI\222\30S\34\b\200\227""0c\234\31r\230r_\226""7]\14*g\236\7[gK\31\7va\7`\213MI\227""6\t4˩U72UJR*Jj\25Xer\225!)\3`۾\205\25#]N\25V\213\177Zd./ahS&\6\14\200\235c\220\17\3\234آ{\rys\213\35\7c\216t\30}Z\6z\235m\210K/nL[*\203sT=\226m\31es\31\207\207l\226K}{]4UysͥmZ<p)A\37""8I>\7Ͽ'\200V\21\201\b&\202\13\6\4\0;";

static constant ARROW_LEFT = "GIF89a\30\0\30\0\221\0\0\0\0\0\0\0!\4\1\0\0\2\0,\0\0\0\0\30\0\30\0@\2+\224\217\33\2\234\213v/}Fj\b&ʶ.ĩhZh\227\206\17\24\0\0;";

static constant ARROW_LEFT_NONACTIVE = "GIF87a\30\0\30\0\0\0\0\0\0,\0\0\0\0\30\0\30\0\0\2\26\204\217\17\234\213޼\17\206H\226S\1\0;";

static constant ARROW_RIGHT = "GIF89a\30\0\30\0\221\0\0\0\0\0\0\0!\4\1\0\0\2\0,\0\0\0\0\30\0\30\0@\2*\224\217\31ߢ\201\207\17^\13Gn\204\235ʶnة\31R֩`r\14\177(\0\0;";

static constant ARROW_RIGHT_NONACTIVE = ARROW_LEFT_NONACTIVE;

static constant ARROW_UP = "GIF89a\30\0\30\0\0\0\0\0\0!\4\1\0\0\1\0,\0\0\0\0\30\0\30\0\0\2+\214\217\0\14\221s\21Yս}_ad\235\222\32i\33ҧM\17\14\n\0;";

static constant ARROW_UP_NONACTIVE = ARROW_LEFT_NONACTIVE;

static constant DROPSHADOW_CORNER = "\211PNG\r\n\32\n\0\0\0\rIHDR\0\0\0\32\0\0\0\32\b\2\0\0\0&(\231\0\0\2""1IDATx\223k\32\1\34\205OmUUF\24<4APA\211A\20*A\211\27\205\200\7\30\227\34Q\211\6&.FZj:ִc\227\14߼~o\204|\226\227\177\224G\20\17>/i\217[KXc\211 \21""1\233ͦ\207\205o'a\224\b\202³dr}}=\32\215-v\13\t\202\205#\21\4e8\34zNnl!6k<\36\3\2qqqq~~h4ji\211/2\21`v\1JR.\227\13\205B>\237\177\215\22V\213\214\202l6jT\215l6\233dt2\231L$\22/6JXY#&}\221j:99rT*\26\213Ep8|tt\24\n\205\236o\22Gk5\"cX\204\216\217\217y)\20\b|Cl\204UR6\214\210/X`\\.n\236Jk\211\23\217@RxzzJ_x\234h^tBZfd2>\221\r\2168\2\233\216tO_d\27v,\26\213h:\235~;{ʾ\4w\214D\"~p\13\226`jjZR=\222\r\216_\204;U&#\27\31hx[T*\24\212\207ZǱl(\200(\213\214\230\202%\227e2\3i\206\23qL\4\348p\205""5|\211\22217v\27\35\203#\213Ⱥ';\35\177\3?\b\205m@4\3\0\0\0\0IENDB`\202";

static Image.image corner1 = Image.PNG.decode(DROPSHADOW_CORNER);
static Image.image corner2 = corner1->rotate_cw();
static Image.image corner3 = corner2->rotate_cw();
static Image.image corner4 = corner3->rotate_cw();


/**
 * HTML and CSS.
 */
static constant PAGE_HEADER =
#"<?xml version=\"1.0\" encoding=\"iso-8859-1\"?>
<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Strict//EN\"
\"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd\">
<html>
<head>
 <aphoto-link type=\"stylesheet\"/>\n";

static constant PAGE_FOOTER =
#"</html>\n";

static constant PREVIOUS_NEXT_ARROWS_HTML =
#"
    <aphoto-link type=\"previous\">
     <aphoto-arrow-left class=\"dirlink\" alt=\"&lt;&lt;\"/>
    </aphoto-link><else><aphoto-arrow-left class=\"dirlink\""
+ #" active=\"false\""
+ #" alt=\"&lt;&lt;\"/></else>
    <aphoto-link type=\"parent\">
     <aphoto-arrow-up  class=\"dirlink\" alt=\"^^\"/>
    </aphoto-link><else><aphoto-arrow-up class=\"dirlink\""
+ #" active=\"false\" alt=\"^^\"/></else>
    <aphoto-link type=\"next\">
     <aphoto-arrow-right class=\"dirlink\" alt=\"&gt;&gt;\"/>
    </aphoto-link><else><aphoto-arrow-right class=\"dirlink\""
+ #" active=\"false\" alt=\"&gt;&gt;\"/></else>";

static constant BOTTOM_ROW_EDIT_HTML =
#"  <tr>
   <td id=\"bottomleft\">
    <aphoto-link type=\"exit\"><i>&lt;exit&nbsp;edit&gt;</i></aphoto-link>
   </td>
   <td id=\"bottomright\"><cset variable=\"var.time\"
       preparse=\"preparse\"><aphoto-date/> <aphoto-time/></cset>
    <input name=\"new_date\" size=\"32\" value=\"&var.time;\"/>
   </td>
  </tr>";

static constant BOTTOM_ROW_HTML =
#"  <tr>
   <td id=\"bottomleft\">
    <i><a href=\"http://www.lysator.liu.se/~norrby/aphoto.html\">aphoto</a></i>"
+ #"
   </td>
   <td id=\"bottomright\"><aphoto-date/>&#8212;" +
"<aphoto-link type=\"edit\"><i>&lt;edit&gt;</i></aphoto-link></td></tr>";

static constant DEFAULT_BROWSING_PAGE =
PAGE_HEADER +
#" <title><aphoto-gallery/><aphoto-title prepend=\": \"/></title>
</head>
<body>
 <table class=\"outer\">
  <tr>
   <td id=\"topleft\">" + PREVIOUS_NEXT_ARROWS_HTML + #"
   </td>
   <td id=\"topright\"><h1><aphoto-title/></h1></td>
  </tr>
  <tr>
   <td id=\"secondhighestleft\"></td>
   <td id=\"secondhighestright\"><i><aphoto-description/></i></td>
  </tr>
 </table>
 <div>
  <aphoto-link type=\"medium\">
   <aphoto-browsing-image id=\"photo\"/>
  </aphoto-link>
 </div>
 <table class=\"outer\">
" + BOTTOM_ROW_HTML + #"
 </table>
</body>
" + PAGE_FOOTER;

static constant DEFAULT_BROWSING_EDIT_PAGE =
PAGE_HEADER +
#" <title><aphoto-gallery/><aphoto-title prepend=\": \"/> (edit)</title>
</head>
<body>
 <aphoto-form method=\"post\">
 <input type=\"hidden\" name=\"aphoto_editing_data\" value=\"posted\" />
 <table class=\"outer\">
  <tr>
   <td id=\"topleft\">" + PREVIOUS_NEXT_ARROWS_HTML + #"
   </td><cset variable=\"var.title\" preparse=\"preparse\"><aphoto-title/>"
+ #"</cset>
   <td id=\"topright\"><h1><aphoto-gallery/>&nbsp;&#8212;&nbsp;"
"<input name=\"new_title\" "
+ #"value=\"&var.title;\"/></h1></td>
  </tr>
  <tr>
   <td id=\"secondhighestleft\">
    Transform&nbsp;(lossless)<br />
    <input type=\"radio\" name=\"trans\" value=\"nop\" checked=\"checked\"/>"
+#"<i>no transf.</i><br />
    <input type=\"radio\" name=\"trans\" value=\"rot90\"/>Rot 90 (CW)<br />
    <input type=\"radio\" name=\"trans\" value=\"rot180\"/>Rot 180<br />
    <input type=\"radio\" name=\"trans\" value=\"rot270\"/>Rot -90 (CW)<br />
    <input type=\"radio\" name=\"trans\" value=\"flipud\"/>Flip U/D<br />
    <input type=\"radio\" name=\"trans\" value=\"fliplr\"/>Flip R/L<br />
   </td>
   <td id=\"secondhighestright\">"
+ #"<cset variable=\"var.desc\""
+ #" preparse=\"preparse\"><aphoto-description/></cset>
    <textarea name=\"new_long_desc\" cols=\"64\" rows=\"4\">" +
#"&var.desc:none;</textarea>
   </td>
  </tr>
  <tr>
   <td id=\"middleleft\"></td>
   <td id=\"middleright\">&nbsp;
    <input type=\"submit\" name=\"submit_and_previous\" value=\"&lt;&#8212;"
+ #"OK\" />
    <input type=\"submit\" name=\"submit_and_stay\" value=\"OK\" />
    <input type=\"submit\" name=\"submit_and_next\" value=\"OK&#8212;&gt;\" />
    <if hidden=\"hidden\"><input type=\"submit\" name=\"unhide\" "
#"value=\"Make visible\" />
    </if><else><input type=\"submit\" name=\"hide_and_next\" "
"value=\"Hide&#8212;&gt;\" "
#"/></else>
   </td>
 </table>
 <div>
  <aphoto-link type=\"medium\">
   <aphoto-browsing-image id=\"photo\"/>
  </aphoto-link>
 </div>
 <table class=\"outer\">
" + BOTTOM_ROW_EDIT_HTML + #"
 </table>
 </aphoto-form>
</body>
" + PAGE_FOOTER;

static constant DEFAULT_FOLDER_PAGE =
PAGE_HEADER +
#"<title>
   <aphoto-gallery/><aphoto-title fallback=\"path\" prepend=\": \"/>
  </title>
</head>
<body>
 <table class=\"outer\">
  <tr>
   <td id=\"topleft\">" + PREVIOUS_NEXT_ARROWS_HTML + #"
   </td>
   <td id=\"topright\">
    <h1><aphoto-gallery/><aphoto-title
                          fallback=\"path\"
                          prepend=\"&amp;nbsp;&amp;#8212;&amp;nbsp;\"/></h1>
   </td>
  </tr>
  <tr>
   <td id=\"secondhighestleft\"></td>
   <td id=\"secondhighestright\">" + FOLDER_INFO_HTML + #"
   </td>
  </tr>
  <tr>
   <td id=\"middleleft\"></td>
   <td id=\"middleright\">" + FOLDERS_HTML + #"
   </td>
  </tr>
  <tr>
   <td id=\"secondlowestleft\"></td>
   <td id=\"secondlowestright\">" + THUMBNAILS_HTML + #"
   </td>
  </tr>
" + BOTTOM_ROW_HTML + #"
 </table>
</body>
" + PAGE_FOOTER;

static constant DEFAULT_FOLDER_EDIT_PAGE =
PAGE_HEADER +
#" <title><aphoto-gallery/><aphoto-title fallback=\"path\" prepend=\": \"/>
   </title>
</head>
<body>
 <aphoto-form method=\"post\">
 <input type=\"hidden\" name=\"aphoto_editing_data\" value=\"posted\" />
 <table class=\"outer\">
  <tr>
   <td id=\"topleft\">" + PREVIOUS_NEXT_ARROWS_HTML + #"
   </td>
   <cset variable=\"var.title\"
         preparse=\"preparse\"><aphoto-title fallback=\"path\"/></cset>
   <td id=\"topright\"><h1><aphoto-gallery/>&nbsp;&#8212;&nbsp;"
"<input name=\"new_title\" " +
#"value=\"&var.title;\"/></h1></td>
  </tr>
  <tr>
   <td id=\"secondhighestleft\"></td>
   <td id=\"secondhighestright\">"
+ #"<cset variable=\"var.desc\""
+ #" preparse=\"preparse\"><aphoto-description/></cset>
     <textarea name=\"new_long_desc\" cols=\"64\" rows=\"4\">" +
#"&var.desc:none;</textarea>
   </td>
  </tr>
  <tr>
   <td id=\"middleleft\">
    <input type=\"submit\" name=\"submit_dir\" value=\"OK\" />
   </td>
   <td id=\"middleright\">" + FOLDERS_HTML + #"
   </td>
  </tr>
  <tr>
   <td id=\"secondlowestleft\"></td>
   <td id=\"secondlowestright\">" + THUMBNAILS_HTML + #"
   </td>
  </tr>
" + BOTTOM_ROW_EDIT_HTML + #"
 </table>
 </aphoto-form>
</body>
" + PAGE_FOOTER;

static constant THUMBNAILS_HTML =
#"<emit source=\"thumbnails\">
    <a href=\"&_.link;\"><imgb0 class=\"thumbnail\" src=\"&_.thumbnail;\" "
"alt=\"\" width=\"&_.width;\" height=\"&_.height;\" title=\"&_.popup_desc;\" />"
"</a></emit>";

static constant FOLDERS_HTML =
#"<emit source=\"folders\">
    <div class=\"folder\"><span><a class=\"folderlink\" href=\"&_.link;\">" +
"<imgb0 class=\"folderlink\" src=\"&_.image;\" height=\"&_.height;\" "
"width=\"&_.width;\" alt=\"[*]\" /></a></span>&nbsp;<a class=\"folderlink\" "
"href=\"&_.link;\">&_.title;</a></div></emit>";

static constant FOLDER_INFO_HTML =
#"    <div><i><autoformat><aphoto-description/></autoformat></i></div>";


static constant APHOTO_CSS = #"
body {
  margin: 0;
  padding: 0;
  border: none;
  background-color: white;
  color: black;
  font-weight: normal;
}
h1 {
  margin: 0;
  padding: 0.1em 0.1em 0.1em 10px;;
  font-family: sans-serif;
  font-weight: normal;
}
div.folder {
  padding: 0px 7px 0px 0px;
  margin: 0 0 0 1em;
}
img.folderlink, .dirlink {
  vertical-align: middle;
  border-style: none;
}
input, textarea {
  background: lightblue;
  border-style: none;
  font-size: 100%;
}
table.outer {
  width: 100%;
  border-style: none;
  border-collapse: collapse;
}
img.thumbnail {
  border-style: none;
}
td#topleft {
  width: 110px;
  padding-left: 1em;
  padding-right: 0.5em;
  border-right: solid black 1px;
  border-bottom: solid black 1px;
  text-align: center;
  vertical-align: middle;
}
td#topright {
  border-left: solid black 1px;
  border-bottom: solid black 1px;
  vertical-align: middle;
}
td#secondhighestleft {
  width: 110px;
  padding-left: 1em;
  padding-right: 0.5em;
  border-right: solid black 1px;
  text-align: left;
  height: 10px;
}
td#secondhighestright {
  border-left: solid black 1px;
  padding: 0.5em 1em 0.2em 15px;
  margin-bottom: 1em;
  height: 10px;
  vertical-align: top;
}
td#middleright {
  border-left: solid black 1px;
  padding-left: 5px;
  vertical-align: top;
}
td#middleleft {
  width: 110px;
  padding-left: 1em;
  padding-right: 0.5em;
  border-right: solid black 1px;
  vertical-align: top;
  text-align: center;
}
td#secondlowestright {
  border-left: solid black 1px;
  padding: 0.5em 0 0.5em 0.2em;
  margin-top: 1em;
}
td#secondlowestleft {
  width: 110px;
  padding-left: 1em;
  padding-right: 0.5em;
  text-align: center;
}
td#bottomright {
  border-left: solid black 1px;
  padding: 0.5em 0 0.3em 1em;
  vertical-align: bottom;
}
td#bottomleft {
  width: 110px;
  padding: 0.5em 0.5em 0.3em 1em;
  border-right: solid black 1px;
  text-align: center;
  vertical-align: bottom;
}
img#photo {
  border: none;
  padding-left: 10px;
}
";

TAGDOCUMENTATION;
constant ARROW_DESCRIPTION=
  "Produces a link to an internal arrow image pointing ";
constant ACTIVE_ARROW_DESCRIPTION=
  #"<attr name='active' value='true|false'>If the returned image link should "
+ #"point
    to an arrow or to a blank image of the same size of the arrow.

    If the value
    of this attribute is 'false', the returned image link will point to a
    blank image.</attr>";

#ifdef manual
constant tagdoc=(["aphoto-arrow-left":#"<desc type='tag'>"
+ ARROW_DESCRIPTION + "to the left</desc>"
+ ACTIVE_ARROW_DESCRIPTION,

"aphoto-arrow-up":#"<desc type='tag'>"
+ ARROW_DESCRIPTION + "upward.</desc>"
+ ACTIVE_ARROW_DESCRIPTION,

"aphoto-arrow-right":#"<desc type='tag'>"
+ ARROW_DESCRIPTION + "to the right.</desc>"
+ ACTIVE_ARROW_DESCRIPTION,

"aphoto-browsing-image":#"<desc type='tag'>"
#"Makes an <b>img</b> tag with source corresponding to a generated image
file pointed to by the current URL. The generated file resides in the
image cache and is meant to be used as a browsing image with a more
convenient size than the original.

For example:

<ex-box>
  <aphoto-browsing-image class='image'/>
</ex-box>

renders an img tag looking somewhat like:

<ex-box>
  <img class='image' height='530' src='/_internal/aphoto!0/xe5dsxlkd7rh56rwr'"
+ #" width='690' />
</ex-box>

All attributes will be passed on to the resulting <b>img</b> tag. Note
the <b>class='image'</b> style argument above for example.</desc>",

"aphoto-date":#"<desc type='tag'>"
+ #"Renders a date string using the current time of the image pointed to
by the current URL.</desc>",

"aphoto-description":#"<desc type='tag'>"
+ #"Echoes a long description associated with the image pointed to
by the current URL.</desc>",

"aphoto-gallery":#"<desc type='tag'>"
+ #"Echoes the value of the 'Gallery name' setting in this config interface
</desc>",

"aphoto-time":#"<desc type='tag'>"
+ #"Renders a time string using the current time of the image pointed to
by the current URL.</desc>",

"aphoto-title":#"<desc type='tag'>"
+ #"Echoes the title of the folder page or image pointed to by the current URL.
<attr name='fallback' value='path'>If the title is empty or non existent
and the value of this attribute is 'path', the title will default to the
image path in the virtual server.</attr>
<attr name='prepend' value='string'>String to prepend to the title if a
title is present. Could be used to prepend a dash.
<ex-box>
  Image<aphoto-title prepend=' -- '/>
</ex-box>

could render a title looking like this:

<ex-box>
  Image -- A giant fish
</ex-box>

or, if the title tag is used in a folder page, the prepend
attribute could be used not to present a dash if at the
root level.
<ex-box>
  Images<aphoto-title fallback='path' prepend=' -- '/>
</ex-box>

could render a title looking like this if not at root level:

<ex-box>
  Images -- path/to/fish
</ex-box>

or like this if at root level:

<ex-box>
  Images
</ex-box>

</attr>
</desc>",

"imgb0":"<desc type='tag'>"
+ #"Works exactly like the img tag, but makes sure that a border='0'
attribute is added if the client is an ancient Netscape
Navigator.</desc>",

"aphoto-form":"<desc type='tag'>"
+ #"Makes an ordinary HTML-form with the correct action according
to the current URL. Usable for data editing pages.</desc>",

"aphoto-link":"<desc type='tag'/>"
+ #"Makes a link of given type.
<attr name='type' value='edit|exit|medium|stylesheet"
+ # "|next|previous|parent'>Defines which type of link to create.
'edit' creates an edit link to the current page, while 'exit'
cxreates a non edit link to the current page. 'medium' creates
a direct link to the current medium pointed to by the current URL.
'stylesheet' creates a link to the built in aphoto stylesheet and
the 'previous', 'next', and 'parent' creates links to neighbouring
media.

<xtable>
 <row>
  <c><p><i>type=edit</i></p></c>
  <c>
   <ex-box><aphoto-link type='edit'>Edit page</aphoto-link></ex-box>
   will render a link to edit the current page, like:
   <ex-box><a href=\"./DSCF0009.JPG.aphoto?edit=true\">Edit page</a></ex-box> 
  </c>
 </row>
 <row>
  <c><p><i>type=exit</i></p></c>
  <c>
   <ex-box><aphoto-link type='exit'>Stop editing</aphoto-link></ex-box>
   will render a link to cease editing the current page, like:
   <ex-box><a href=\"./DSCF0009.JPG.aphoto\">Stop editing</a></ex-box> 
  </c>
 </row>
 <row>
  <c><p><i>type=medium</i></p></c>
  <c>
<ex-box>
<aphoto-link type='medium'>
 <aphoto-browsing-image/>
</aphoto-link>
</ex-box>
   will render a link to real medium file,\n
   (but not line broken as it is in the example box below):
<ex-box>
<a href=\"./DSCF0009.JPG\">
 <img height='530' width='690' 
      src='/_internal/aphoto!0/xe5dsxlkd7rh56rwr' />
</a>
</ex-box> 
  </c>
 </row>
 <row>
  <c><p><i>type=stylesheet</i></p></c>
  <c>
   <ex-box><aphoto-link type='stylesheet'/></ex-box>
   will render a stylesheet link, like this (but not line broken\n
   as it is in the example box below):
<ex-box>
<link href=\"/_internal/aphoto2!0/aphoto.css\" type=\"text/css\"
      title=\"default aphoto style\" rel=\"stylesheet\" />
</ex-box> 
  </c>
 </row>
 <row>
  <c><p><i>type={previous|next}</i></p></c>
  <c>
   The previous link for example,
   <ex-box><aphoto-link type='previous'>Previous image</aphoto-link></ex-box>
   will render a link to an adjacent medium file:
   <ex-box><a href=\"./DSCF0008.JPG.aphoto\">Previous image</a></ex-box> 
  </c>
 </row>
 <row>
  <c><p><i>type=parent</i></p></c>
  <c>Renders a link to the parent directory.</c>
 </row>
</xtable>
</attr></desc>",

//----------------------------------------------------------------------

"if#hidden":#"<desc type='plugin'><short>
This if plugin returns true if the medium pointed to by the current
URL is hidden</short>
<attr name='hidden' value='hidden'>
This attribute must be present for the hidden test to take place.
</attr>
<ex-box>
<if hidden='hidden'>
 Current image is hidden.
</if>
</else>
 Current image is visible.
<else>
</ex-box>
</desc>
",

//----------------------------------------------------------------------

"emit#folders":({ #"<desc type='plugin'><short>
 Prints available folders.</short> This plugin makes it easy to list all
 folders in a directory and also print their aphoto properties, like
 long and short descriptions.</desc>",
                  ([
"&_.link;":#"<desc type='entity'>A link to the folder suitable for placing
as <b>href</b> in anchor tags.</desc>",
"&_.long_desc;":#"<desc type='entity'>The long description connected
to the folder</desc>",
"&_.image;":#"<desc type='entity'>An internal image to use as folder
image when creating links to the folder.</desc>",
"&_.width;":#"<desc type='entity'>The width of the internal folder image.
</desc>",
"&_.height;":#"<desc type='entity'>The height of the internal folder image.
</desc>",
"&_.title;":#"<desc type='entity'>The short descriptive title of the
folder.</desc>",

		])
	     }),
"emit#thumbnails":({ #"<desc type='plugin'><short>
 Prints available thumbnails.</short> This plugin makes it easy to list all
 images in a directory and also print their aphoto properties, like
 long and short descriptions and cached thumbnail images.</desc>",
                  ([
"&_.link;":#"<desc type='entity'>A link to the medium in question,
suitable for placing as <b>href</b> in anchor tags.</desc>",
"&_.long_desc;":#"<desc type='entity'>The long description connected
to the medium</desc>",
"&_.popup_desc;":#"<desc type='entity'>A portion of the long description,
suitable for popups when hoovering over a thumbnail.</desc>",
"&_.width;":#"<desc type='entity'>The width of the thumbnail image.</desc>",
"&_.height;":#"<desc type='entity'>The height of the thumbnail image.</desc>",
"&_.title;":#"<desc type='entity'>The short descriptive title of the
medium pointed to by the thumbnail.</desc>",

		])
	     }),
]);
#endif
