/****************************************************************************
 This is the Automatic Photos module for Roxen WebServer 4.0--
 Copyright (C) 2001--2005 -- M Norrby <norrby@lysator.liu.se>

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.         
 
 To acquire the latest version of aphoto, go to
 <URL:http://www.lysator.liu.se/~norrby/aphoto.html>.
 
 Please send comments and bug-reports to <norrby+aphoto@lysator.liu.se>.
****************************************************************************/

#include <module.h>
#include <request_trace.h>

inherit "module";

/**
 * Module constants. Read by the server when loading the module.
 */
constant cvs_version = "$Id: aphoto_directories.pike,v 1.6 2005/01/16 16:44:53 norrby Exp $";
int module_type = MODULE_DIRECTORIES;
constant module_name = "Aphoto directories";
constant module_unique = 1;
constant thread_safe = 1;
constant module_doc = "";


/**
 * Module private constants.
 */
static constant SETTING_INDEX_FILES = "index-files";


/**
 * Implemented module API functions in alphabetic order.
 */
void
create(Configuration conf)
{
  defvar(SETTING_INDEX_FILES,
	 Variable.StringList(({"index.xml",
			       "index.html",
			       "index.pike",}),
			     VAR_INITIAL,
			     "Index files",
                             "If one of these files is present in a "
                             "directory, it will be returned "
                             "instead of the folder page."));
}

mapping
parse_directory(RequestID id)
{
  if (should_relocate_and_add_slash(id->not_query))
    return Roxen.http_redirect(id->not_query + "/", id);

  if (id->not_query[-1] == '/')
  {
    foreach(query(SETTING_INDEX_FILES) - ({""}), string file) {
      if (file_exists(file, id))
      {
        mapping handled_request = get_index_file(file, id);
        if (!!handled_request)
          return handled_request;
      }
    }
  }

  if (file_exists(".aphoto", id))
    id->variables->aphoto_warning_old_aphoto_file_exists = 1;
  
  return (id->conf->find_module("aphoto")
          ||id->conf->find_module("aphoto2"))->parse_directory(id);
}

mapping(LocaleString:function(RequestID:void))
query_action_buttons(RequestID id)
{
  return ([]);
}

void
start(int num, Configuration config)
{
  //removed until thte problem with both old and new aphoto
  //together is resolved
  //  module_dependencies(config, ({"aphoto"}));
}
  
string
status()
{
}


/**
 * Aphoto directories: internal functions in alphabetic order.
 */
mapping get_index_file(string file, RequestID id)
{
  id->not_query += file;
  mixed ret = id->conf->handle_request(id);
  if (ret && mappingp(ret))
    return ret;
  else
    return 0;
}

static int(0..1)
file_exists(string file, RequestID id)
{
  array stat;
  return (stat = id->conf->stat_file(id->not_query + file, id))
    && (stat[ST_SIZE] >= 0);
}

static int(0..1)
should_relocate_and_add_slash(string not_query)
{
  return not_query == "" || (not_query[-1] != '/' && not_query[-1] != '.');
}
